"""
All usage of SQLAlchemy should be imported via this module to allow SQLAlchemy to be
installed optionally.

Sufficient interface is implemented to permit import of SQLAlchemy based classes without
any error. If any attempt is made to use those classes a meaningful warning is returned.
"""

try:
    import sqlalchemy as sa
    from sqlalchemy.ext import declarative
except ImportError:
    class MockSQlAlchemy:
        def __getattr__(self, item):
            return self

        def __call__(self, *args, **kwargs):
            return self

        def __mro_entries__(self, *args, **kwargs):
            return tuple()

        def declarative_base(self):
            return MockBase

        def __getitem__(self, item):
            fail()

        def __setitem__(self, key, value):
            fail()


    sa = MockSQlAlchemy()
    declarative = sa


    def fail():
        raise ImportError(
            "Please install SQLAlchemy to use the database"
        )


    class MockBase:
        def __init__(self, *args, **kwargs):
            fail()

        metadata = sa



from typing import Optional

from autoconf import conf
from autofit.database.sqlalchemy_ import sa
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import (
    InitializerPrior,
    AbstractInitializer,
    SpecificRangeInitializer,
)
from autofit.tools.util import IntervalCounter


class AbstractNest(NonLinearSearch):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        prior_passer: Optional[PriorPasser] = None,
        iterations_per_update: Optional[int] = None,
        number_of_cores: Optional[int] = None,
        session: Optional[sa.orm.Session] = None,
        initializer: Optional[AbstractInitializer] = None,
        **kwargs
    ):
        """
        Abstract class of a nested sampling `NonLinearSearch` (e.g. MultiNest, Dynesty).

        **PyAutoFit** allows a nested sampler to automatically terminate when the acceptance ratio falls below an input
        threshold value. When this occurs, all samples are accepted using the current maximum log likelihood value,
        irrespective of how well the model actually fits the data.

        This feature should be used for non-linear searches where the nested sampler gets 'stuck', for example because
        the log likelihood function is stochastic or varies rapidly over small scales in parameter space. The results of
        samples using this feature are not realiable (given the log likelihood is being manipulated to end the run), but
        they are still valid results for linking priors to a new search and non-linear search.

        Parameters
        ----------
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        session
            An SQLAlchemy session instance so the results of the model-fit are written to an SQLite database.
        """
        if isinstance(initializer, SpecificRangeInitializer):
            raise ValueError(
                "SpecificRangeInitializer cannot be used for nested sampling"
            )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer or InitializerPrior(),
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

    class Fitness(NonLinearSearch.Fitness):
        def __init__(
            self, paths, analysis, model, samples_from_model, log_likelihood_cap=None
        ):
            super().__init__(
                paths=paths,
                analysis=analysis,
                model=model,
                samples_from_model=samples_from_model,
                log_likelihood_cap=log_likelihood_cap,
            )

            self.stagger_accepted_samples = 0
            self.resampling_figure_of_merit = -1.0e99

            self.should_check_terminate = IntervalCounter(1000)

        def figure_of_merit_from(self, parameter_list):
            """The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space. All Nested
            samplers use the log likelihood.
            """
            return self.log_likelihood_from(parameter_list=parameter_list)

    @property
    def config_type(self):
        return conf.instance["non_linear"]["nest"]

    def fitness_function_from_model_and_analysis(
        self, model, analysis, log_likelihood_cap=None
    ):

        return self.__class__.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap,
        )



from typing import Optional, Dict, Tuple, Any, Union

from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.expectation_propagation.optimiser import AbstractFactorOptimiser
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.laplace import newton
from autofit.graphical.mean_field import MeanField, FactorApproximation
from autofit.graphical.utils import Status
from autofit.mapper.variable_operator import VariableData

FactorApprox = Union[EPMeanField, FactorApproximation, Factor]


def make_posdef_hessian(mean_field, variables):
    return MeanField.precision(mean_field, variables)


class LaplaceOptimiser(AbstractFactorOptimiser):
    def __init__(
        self,
        make_hessian=make_posdef_hessian,
        search_direction=newton.newton_abs_direction,
        calc_line_search=newton.line_search,
        quasi_newton_update=newton.full_diag_update,
        stop_conditions=newton.stop_conditions,
        make_det_hessian=None,
        max_iter: int = 100,
        n_refine: int = 3,
        hessian_kws: Optional[Dict[str, Any]] = None,
        det_hessian_kws: Optional[Dict[str, Any]] = None,
        search_direction_kws: Optional[Dict[str, Any]] = None,
        line_search_kws: Optional[Dict[str, Any]] = None,
        quasi_newton_kws: Optional[Dict[str, Any]] = None,
        stop_kws: Optional[Dict[str, Any]] = None,
        check_limits=True,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.make_hessian = make_hessian
        self.make_det_hessian = make_det_hessian or make_hessian
        self.search_direction = search_direction
        self.calc_line_search = calc_line_search
        self.quasi_newton_update = quasi_newton_update
        self.stop_conditions = stop_conditions

        self.max_iter = max_iter
        self.n_refine = n_refine

        self.hessian_kws = hessian_kws or {}
        self.det_hessian_kws = det_hessian_kws or hessian_kws or {}
        self.search_direction_kws = search_direction_kws or {}
        self.line_search_kws = line_search_kws or {}
        self.quasi_newton_kws = quasi_newton_kws or {}
        self.stop_kws = stop_kws or {}
        self.check_limits = check_limits

    @property
    def default_kws(self):
        return dict(
            max_iter=self.max_iter,
            n_refine=self.n_refine,
            search_direction=self.search_direction,
            calc_line_search=self.calc_line_search,
            quasi_newton_update=self.quasi_newton_update,
            stop_conditions=self.stop_conditions,
            search_direction_kws=self.search_direction_kws,
            line_search_kws=self.line_search_kws,
            quasi_newton_kws=self.quasi_newton_kws,
            stop_kws=self.stop_kws,
        )

    def prepare_state(
        self,
        factor_approx: FactorApprox,
        mean_field: MeanField = None,
        params: VariableData = None,
    ) -> newton.OptimisationState:
        mean_field = mean_field or factor_approx.model_dist

        free_variables = factor_approx.free_variables
        det_variables = factor_approx.deterministic_variables

        parameters = MeanField.mean.fget(mean_field)
        if params:
            for v, p in params.items():
                parameters[v] = p

        hessian = self.make_hessian(mean_field, free_variables, **self.hessian_kws)
        if det_variables:
            det_hessian = self.make_hessian(mean_field, det_variables)
        else:
            det_hessian = None

        kws = {}
        if self.check_limits:
            kws["upper_limit"] = MeanField.upper_limit.fget(mean_field)
            kws["lower_limit"] = MeanField.lower_limit.fget(mean_field)

        return newton.OptimisationState(
            factor_approx,
            factor_approx.func_gradient,
            parameters.subset(free_variables),
            hessian,
            det_hessian,
            **kws
        )

    def optimise_state(
        self,
        state: newton.OptimisationState,
        old_state: Optional[newton.OptimisationState] = None,
        **kwargs
    ) -> Tuple[bool, newton.OptimisationState, str]:
        kws = {**self.default_kws, **kwargs}
        return newton.optimise_quasi_newton(state, old_state, **kws)

    def optimise_approx(
        self,
        factor_approx: FactorApprox,
        mean_field: MeanField = None,
        params: VariableData = None,
        **kwargs
    ) -> Tuple[MeanField, Status]:

        mean_field = mean_field or factor_approx.model_dist
        state = self.prepare_state(factor_approx, mean_field, params)
        next_state, status = self.optimise_state(state, **kwargs)
        # if status.flag != StatusFlag.SUCCESS:
        next_state = max(state, next_state, key=lambda x: x.value)
        next_state = self.refine_state(
            next_state, mean_field.sample, n_refine=kwargs.get("n_refine")
        )

        projection = mean_field.from_opt_state(next_state)
        return projection, status

    def refine_state(self, state, new_param, n_refine=None):
        next_state = state
        for i in range(n_refine or self.n_refine):
            new_state = state.update(parameters=new_param())
            next_state = self.quasi_newton_update(
                next_state, new_state, **self.quasi_newton_kws
            )

        return next_state

    def refine_approx(
        self,
        factor_approx: FactorApprox,
        mean_field: MeanField = None,
        params: VariableData = None,
        n_refine=None,
    ) -> Tuple[MeanField, Status]:
        mean_field = mean_field or factor_approx.model_dist
        state = self.prepare_state(factor_approx, mean_field, params)
        next_state = self.refine_state(state, mean_field.sample, n_refine=n_refine)
        return mean_field.from_opt_state(next_state)

    def optimise(
        self,
        factor_approx: FactorApproximation,
        status: Optional[Status] = Status(),
        **kwargs
    ) -> Tuple[MeanField, Status]:
        return self.optimise_approx(factor_approx, **kwargs)



from abc import ABC, abstractmethod
import numpy as np
from os import path

from autofit import exc
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples import Samples
from autofit.text import text_util

class Placeholder:
    def __getattr__(self, item):
        """
        Placeholders return None to represent the missing result's value
        """
        return None

    def __getstate__(self):
        return {}

    def __setstate__(self, state):
        pass

    def __gt__(self, other):
        return False

    def __lt__(self, other):
        return True

    @property
    def samples(self):
        return self


class AbstractResult(ABC):
    def __init__(self, sigma):
        self._instance = None
        self.sigma = sigma

    @property
    @abstractmethod
    def samples(self):
        pass

    @property
    @abstractmethod
    def model(self):
        pass

    @property
    def info(self) -> str:
        return text_util.result_info_from(
            samples=self.samples,
        )

    def __gt__(self, other):
        """
        Results are sorted by their associated log_likelihood.

        Placeholders are always low.
        """
        if isinstance(other, Placeholder):
            return True
        return self.log_likelihood > other.log_likelihood

    def __lt__(self, other):
        """
        Results are sorted by their associated log_likelihood.

        Placeholders are always low.
        """
        if isinstance(other, Placeholder):
            return False
        return self.log_likelihood < other.log_likelihood

    @property
    def log_likelihood(self):
        return max(self.samples.log_likelihood_list)

    @property
    def instance(self):
        if self._instance is None:
            self._instance = self.samples.max_log_likelihood()
        return self._instance

    @property
    def max_log_likelihood_instance(self):
        return self.instance

    def model_absolute(self, a: float) -> AbstractPriorModel:
        """
        Parameters
        ----------
        a
            The absolute width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined absolute
        width.
        """
        return self.model.mapper_from_gaussian_tuples(
            self.samples.gaussian_priors_at_sigma(sigma=self.sigma), a=a
        )

    def model_relative(self, r: float) -> AbstractPriorModel:
        """
        Parameters
        ----------
        r
            The relative width of gaussian priors

        Returns
        -------
        A model mapper created by taking results from this search and creating priors with the defined relative
        width.
        """
        return self.model.mapper_from_gaussian_tuples(
            self.samples.gaussian_priors_at_sigma(sigma=self.sigma), r=r
        )


class Result(AbstractResult):
    def __init__(
            self,
            samples: Samples,
            model,
            sigma=3.0,
            use_errors=True,
            use_widths=True
    ):
        """
        The result of a non-linear search, which includes:

        - The samples of the non-linear search (E.g. MCMC chains, nested sampling samples) which are used to compute
        the maximum likelihood model, posteriors and other properties.

        - The model used to fit the data, which uses the samples to create specific instances of the model (e.g.
        an instance of the maximum log likelihood model).

        - The non-linear search used to perform the model fit.

        Parameters
        ----------
        model
            The model mapper from the stage that produced this result
        """
        super().__init__(sigma)
        if samples is not None:
            samples.model = model

        self._samples = samples

        self.use_errors = use_errors
        self.use_widths = use_widths

        self._model = model
        self.__model = None

        self.child_results = None

    @property
    def samples(self):
        return self._samples

    @property
    def projected_model(self) -> AbstractPriorModel:
        """
        Create a new model with the same structure as the previous model,
        replacing each prior with a new prior created by calculating sufficient
        statistics from samples and corresponding weights for that prior.
        """
        weights = self.samples.weight_list
        arguments = {
            prior: prior.project(
                samples=np.array(
                    self.samples.values_for_path(
                        path
                    )
                ),
                weights=weights,
            )
            for path, prior
            in self._model.path_priors_tuples
        }
        return self._model.mapper_from_prior_arguments(
            arguments
        )

    @property
    def model(self):
        if self.__model is None:
            tuples = self.samples.gaussian_priors_at_sigma(
                sigma=self.sigma
            )
            self.__model = self._model.mapper_from_gaussian_tuples(
                tuples,
                use_errors=self.use_errors,
                use_widths=self.use_widths
            )
        return self.__model

    @model.setter
    def model(self, model):
        self.__model = model

    def __str__(self):
        return "Analysis Result:\n{}".format(
            "\n".join(
                ["{}: {}".format(key, value) for key, value in self.__dict__.items()]
            )
        )

    def __getitem__(self, item):
        return self.child_results[item]

    def __iter__(self):
        return iter(self.child_results)

    def __len__(self):
        return len(self.child_results)


class ResultsCollection:
    def __init__(self, result_list=None):
        """
        A collection of results from previous searches. Results can be obtained using an index or the name of the search
        from whence they came.
        """
        self.__result_list = []
        self.__result_dict = {}

        if result_list is not None:
            for result in result_list:
                self.add(name="", result=result)

    def copy(self):
        collection = ResultsCollection()
        collection.__result_dict = self.__result_dict
        collection.__result_list = self.__result_list
        return collection

    @property
    def reversed(self):
        return reversed(self.__result_list)

    @property
    def last(self):
        """
        The result of the last search
        """
        if len(self.__result_list) > 0:
            return self.__result_list[-1]
        return None

    @property
    def first(self):
        """
        The result of the first search
        """
        if len(self.__result_list) > 0:
            return self.__result_list[0]
        return None

    def add(self, name, result):
        """
        Add the result of a search.

        Parameters
        ----------
        name: str
            The name of the search
        result
            The result of that search
        """
        try:
            self.__result_list[self.__result_list.index(result)] = result
        except ValueError:
            self.__result_list.append(result)
        self.__result_dict[name] = result

    def __getitem__(self, item):
        """
        Get the result of a previous search by index

        Parameters
        ----------
        item: int
            The index of the result

        Returns
        -------
        result: Result
            The result of a previous search
        """
        return self.__result_list[item]

    def __len__(self):
        return len(self.__result_list)

    def from_name(self, name):
        """
        Returns the result of a previous search by its name

        Parameters
        ----------
        name: str
            The name of a previous search

        Returns
        -------
        result: Result
            The result of that search

        Raises
        ------
        exc.PipelineException
            If no search with the expected result is found
        """
        try:
            return self.__result_dict[name]
        except KeyError:
            raise exc.PipelineException(
                "No previous search named {} found in results ({})".format(
                    name, ", ".join(self.__result_dict.keys())
                )
            )

    def __contains__(self, item):
        return item in self.__result_dict



import math
from typing import Optional, Tuple

from autoconf import conf
from autofit import exc
from autofit.graphical import FactorApproximation
from autofit.graphical.utils import Status
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.mock.mock_result import MockResult
from autofit.non_linear.mock.mock_samples import MockSamples
from autofit.non_linear.samples import Sample


def samples_with_log_likelihood_list(
        log_likelihood_list,
        kwargs
):
    if isinstance(log_likelihood_list, float):
        log_likelihood_list = [log_likelihood_list]
    return [
        Sample(
            log_likelihood=log_likelihood,
            log_prior=0,
            weight=0,
            kwargs=kwargs,
        )
        for log_likelihood
        in log_likelihood_list
    ]


def _make_samples(model):
    return {
        path: prior.value_for(0.5)
        for path, prior
        in model.path_priors_tuples
    }


class MockSearch(NonLinearSearch):
    def __init__(
            self,
            name="",
            samples=None,
            result=None,
            unique_tag: Optional[str] = None,
            prior_passer=None,
            fit_fast=True,
            sample_multiplier=1,
            save_for_aggregator=False,
            return_sensitivity_results=False,
            **kwargs
    ):

        super().__init__(name=name, unique_tag=unique_tag, **kwargs)

        self.samples = samples or MockSamples()

        self.result = MockResult(samples=samples) if result is None else result

        if prior_passer is not None:
            self.prior_passer = prior_passer

        self.fit_fast = fit_fast
        self.sample_multiplier = sample_multiplier

        self.save_for_aggregator = save_for_aggregator
        self.return_sensitivity_results = return_sensitivity_results

    def check_model(self, model):
        pass

    @property
    def config_type(self):
        return conf.instance["non_linear"]["mock"]

    @property
    def config_dict_search(self):
        return {}

    def _fit_fast(self, model, analysis, log_likelihood_cap=None):
        class Fitness:
            def __init__(self, instance_from_vector, result):
                self.result = result
                self.instance_from_vector = instance_from_vector

            def __call__(self, vector):
                instance = self.instance_from_vector(vector)

                log_likelihood = analysis.log_likelihood_function(instance)

                if self.result.instance is None:
                    self.result.instance = instance

                # Return Chi squared
                return -2 * log_likelihood

        if self.save_for_aggregator:
            analysis.save_attributes_for_aggregator(paths=self.paths)

        fitness_function = Fitness(model.instance_from_vector, result=self.result)
        fitness_function([prior.mean for prior in model.priors_ordered_by_id])

        return fitness_function.result

    def _fit(self, model, analysis, log_likelihood_cap=None):

        if self.fit_fast:
            result = self._fit_fast(model=model, analysis=analysis)
            return result

        if model.prior_count == 0:
            raise AssertionError("There are no priors associated with the model!")
        if model.prior_count != len(model.unique_prior_paths):
            raise AssertionError(
                "Prior count doesn't match number of unique prior paths"
            )
        index = 0
        unit_vector = model.prior_count * [0.5]
        while True:
            try:
                instance = model.instance_from_unit_vector(unit_vector)
                fit = analysis.log_likelihood_function(instance)
                break
            except exc.FitException as e:
                unit_vector[index] += 0.1
                if unit_vector[index] >= 1:
                    raise e
                index = (index + 1) % model.prior_count
        samples = MockSamples(
            sample_list=samples_with_log_likelihood_list(
                self.sample_multiplier * fit,
                _make_samples(model)
            ),
            model=model,
            gaussian_tuples=[
                (prior.mean, prior.width if math.isfinite(prior.width) else 1.0)
                for prior in sorted(model.priors, key=lambda prior: prior.id)
            ],
        )

        self.paths.save_samples(samples)

        return analysis.make_result(
            model=model,
            samples=samples,
        )

    def perform_update(self, model, analysis, during_analysis):

        if self.samples is not None and not self.return_sensitivity_results:
            self.paths.save_object("samples", self.samples)
            return self.samples

        return MockSamples(
            sample_list=samples_with_log_likelihood_list(
                [1.0, 2.0],
                _make_samples(model)
            ),
            gaussian_tuples=[
                (prior.mean, prior.width if math.isfinite(prior.width) else 1.0)
                for prior in sorted(model.priors, key=lambda prior: prior.id)
            ]
        )


class MockOptimizer(MockSearch):

    def __init__(self, **kwargs):
        super().__init__(fit_fast=False, **kwargs)

    @property
    def samples_cls(self):
        return MockOptimizer

    def project(
            self,
            factor_approx: FactorApproximation,
            status: Status = Status()
    ) -> Tuple[FactorApproximation, Status]:
        pass

    init_args = list()



from . import utils
from .declarative.abstract import PriorFactor
from .declarative.collection import FactorGraphModel
from .declarative.factor.analysis import AnalysisFactor
from .declarative.factor.hierarchical import _HierarchicalFactor, HierarchicalFactor
from .expectation_propagation.ep_mean_field import EPMeanField
from .expectation_propagation.optimiser import EPOptimiser
from .expectation_propagation import StochasticEPOptimiser
from .factor_graphs import FactorGraph
from .factor_graphs.factor import Factor
from .laplace import LaplaceOptimiser, OptimisationState
from .mean_field import FactorApproximation, MeanField
from .utils import Status
from .. import messages
from ..mapper.variable import Variable, Plate, VariableData, FactorValue, variables

dir(Variable)
dir(messages)
dir(FactorValue)



class MockClassx2:
    def __init__(self, one=1, two=2):
        self.one = one
        self.two = two


class MockClassx2Instance(MockClassx2):
    pass


class MockClassx2FormatExp:
    """
    This mock classes's second parameter `two_exp` has a format label `two={:.2e}` in `notational/label_format.ini` and
     is used to test latex generation.
    """

    def __init__(self, one=1, two_exp=2):
        self.one = one
        self.two_exp = two_exp


class MockClassx2NoSuperScript:
    def __init__(self, one=1, two=2):
        self.one = one
        self.two = two


class MockClassx4:
    def __init__(self, one=1, two=2, three=3, four=4):
        self.one = one
        self.two = two
        self.three = three
        self.four = four


class MockClassx3(MockClassx4):
    def __init__(self, one=1, two=2, three=3):
        super().__init__(one, two, three)


class MockClassx2Tuple:
    def __init__(self, one_tuple=(0.0, 0.0)):
        """Abstract MockParent, describing an object with y, x cartesian
        coordinates """
        self.one_tuple = one_tuple

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class MockClassx3TupleFloat:
    def __init__(self, one_tuple=(0.0, 0.0), two=0.1):
        self.one_tuple = one_tuple
        self.two = two


class MockClassRelativeWidth:
    def __init__(self, one, two, three):
        self.one = one
        self.two = two
        self.three = three


class MockClassInf:
    def __init__(self, one, two):
        self.one = one
        self.two = two


class MockComplexClass:
    def __init__(self, simple: MockClassx2):
        self.simple = simple


class MockDeferredClass:
    def __init__(self, one, two):
        self.one = one
        self.two = two


class MockListClass:
    def __init__(self, ls: list):
        self.ls = ls


class MockWithFloat:
    def __init__(self, value):
        self.value = value


class MockWithTuple:
    def __init__(self, tup=(0.0, 0.0)):
        self.tup = tup


class MockOverload:

    def __init__(self, one=1.0):
        self.one = one

    def with_two(self, two):
        self.two = two

    @property
    def two(self):
        return self.one * 2

    @two.setter
    def two(self, two):
        self.one = two / 2


class MockComponents:
    def __init__(
            self,
            components_0: list = None,
            components_1: list = None,
            parameter=None,
            **kwargs
    ):
        self.parameter = parameter
        self.group_0 = components_0
        self.group_1 = components_1
        self.kwargs = kwargs


class MockParent:
    def __init__(self, tup=(0.0, 0.0)):
        self.tup = tup

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class MockChildTuple(MockParent):
    def __init__(self, tup=(0.0, 0.0)):
        """ Generic circular profiles class to contain functions shared by light and
        mass profiles.

        Parameters
        ----------
        tup
            The (y,x) coordinates of the origin of the profile.
        """
        super().__init__(tup)


class MockChildTuplex2(MockChildTuple):
    def __init__(self, tup=(0.0, 0.0), one=1.0, two=0.0):
        """ Generic elliptical profiles class to contain functions shared by light
        and mass profiles.

        Parameters
        ----------
        tup
            The (y,x) coordinates of the origin of the profiles
        one
            Ratio of profiles ellipse's minor and major axes (b/a)
        two : float
            Rotational two of profiles ellipse counter-clockwise from positive x-axis
        """
        super().__init__(tup)
        self.one = one
        self.two = two


class MockChildTuplex3(MockChildTuple):
    def __init__(self, tup=(0.0, 0.0), one=1.0, two=0.0, three=0.0):
        """ Generic elliptical profiles class to contain functions shared by light
        and mass profiles.

        Parameters
        ----------
        tup
            The (y,x) coordinates of the origin of the profiles
        one
            Ratio of profiles ellipse's minor and major axes (b/a)
        two : float
            Rotational two of profiles ellipse counter-clockwise from positive x-axis
        """
        super().__init__(tup)
        self.one = one
        self.two = two
        self.three = three



from typing import Optional, List

from autofit.database.sqlalchemy_ import sa


class SettingsSearch:
    def __init__(
            self,
            path_prefix: str,
            unique_tag: Optional[str] = None,
            number_of_cores: Optional[int] = 1,
            session: Optional[sa.orm.Session] = None,
            info: Optional[dict] = None,
            pickle_files: Optional[List[str]] = None,
    ):
        """
        Stores all the input settings that are used in search's and their `fit functions.

        This is used for more concisely passing settings through pipelines written using the empirical Bayesian
        functionality.

        Parameters
        ----------
        path_prefix
            The prefix of folders between the output path and the search folders.
        unique_tag
            The unique tag for this model-fit, which will be given a unique entry in the sqlite database and also acts as
            the folder after the path prefix and before the search name. This is typically the name of the dataset.
        number_of_cores
            The number of CPU cores used to parallelize the model-fit. This is used internally in a non-linear search
            for most model fits, but is done on a per-fit basis for grid based searches (e.g. sensitivity mapping).
        session
            The SQLite database session which is active means results are directly wrtten to the SQLite database
            at the end of a fit and loaded from the database at the start.
        info
            Optional dictionary containing information about the model-fit that is stored in the database and can be
            loaded by the aggregator after the model-fit is complete.
        pickle_files
            Optional pickle files which are accessible via the database post model-fitting.
        """

        self.path_prefix = path_prefix
        self.unique_tag = unique_tag
        self.number_of_cores = number_of_cores
        self.session = session

        self.info = info
        self.pickle_files = pickle_files

    @property
    def search_dict(self):
        return {
            "path_prefix": self.path_prefix,
            "unique_tag": self.unique_tag,
            "number_of_cores": self.number_of_cores,
            "session": self.session,
        }

    @property
    def search_dict_x1_core(self):
        return {
            "path_prefix": self.path_prefix,
            "unique_tag": self.unique_tag,
            "number_of_cores": 1,
            "session": self.session,
        }

    @property
    def fit_dict(self):
        return {"info": self.info, "pickle_files": self.pickle_files}



from abc import ABC, abstractmethod
from collections import Counter
from typing import Set, List, Dict, Optional
from typing import Tuple

from autofit.graphical.declarative.factor.prior import PriorFactor
from autofit.graphical.declarative.graph import DeclarativeFactorGraph
from autofit.graphical.expectation_propagation import AbstractFactorOptimiser
from autofit.graphical.expectation_propagation import EPMeanField, EPOptimiser
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.collection import Collection
from autofit.mapper.variable import Plate
from autofit.messages.normal import NormalMessage
from autofit.non_linear.analysis import Analysis
from autofit.non_linear.paths.abstract import AbstractPaths


class AbstractDeclarativeFactor(Analysis, ABC):
    optimiser: AbstractFactorOptimiser
    _plates: Tuple[Plate, ...] = ()

    def __init__(self, include_prior_factors=False):
        self.include_prior_factors = include_prior_factors

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    def model_factors(self) -> List["AbstractDeclarativeFactor"]:
        return [self]

    @property
    @abstractmethod
    def prior_model(self):
        pass

    @property
    @abstractmethod
    def info(self):
        pass

    @property
    def priors(self) -> Set[Prior]:
        """
        A set of all priors encompassed by the contained likelihood models
        """
        return {
            prior
            for model
            in self.model_factors
            for prior
            in model.prior_model.priors
        }

    @property
    def prior_counts(self) -> List[Tuple[Prior, int]]:
        """
        A dictionary mapping unique priors in the graph to the number of factors
        in which each prior occurs.
        """
        counter = Counter()
        for factor in self.model_factors:
            for prior in factor.prior_model.priors:
                counter[prior] += 1
        return [
            (prior, count + 1 if self.include_prior_factors else count)
            for prior, count
            in counter.items()
        ]

    @property
    def prior_factors(self) -> List[PriorFactor]:
        """
        A list of factors that act as priors on latent variables. One factor exists
        for each unique prior.
        """
        return list(map(PriorFactor, sorted(self.priors)))

    @property
    def message_dict(self) -> Dict[Prior, NormalMessage]:
        """
        Dictionary mapping priors to messages.

        Messages are computed to the power of one over the number of factors
        a prior is shared by. This means that when the cavity distribution
        is computed for the first optimisation the message has the same sigma
        as the user defined prior.
        """
        return {
            prior: prior.message ** (1 / (count - 1)) if count > 1 else prior.message
            for prior, count
            in self.prior_counts
        }

    @property
    def graph(self) -> DeclarativeFactorGraph:
        """
        The complete graph made by combining all factors and priors
        """
        factors = [
            model
            for model
            in self.model_factors
        ]
        if self.include_prior_factors:
            factors += self.prior_factors
        # noinspection PyTypeChecker
        return DeclarativeFactorGraph(factors)

    def draw_graph(
            self,
            **kwargs
    ):
        """
        Visualise the graph.

        Variables and Factors are nodes labelled according to their corresponding
        prior or model respectively.

        Parameters
        ----------
        kwargs
            Arguments passed to visualisation
        """
        graph = self.graph

        factor_labels = {
            factor: factor.name
            if factor.label is None
            else factor.label
            for factor in graph.factors
        }
        variable_labels = {
            variable: variable.name
            if variable.label is None
            else variable.label
            for variable in graph.all_variables
        }

        import matplotlib.pyplot as plt
        if "draw_labels" not in kwargs:
            kwargs["draw_labels"] = True
        if "variable_labels" not in kwargs:
            kwargs["variable_labels"] = variable_labels
        if "factor_labels" not in kwargs:
            kwargs["factor_labels"] = factor_labels
        graph.draw_graph(
            **kwargs
        )
        plt.show()

    @property
    def plates(self):
        return self._plates

    def mean_field_approximation(self) -> EPMeanField:
        """
        Returns a EPMeanField of the factor graph
        """
        return EPMeanField.from_approx_dists(
            self.graph,
            self.message_dict
        )

    def _make_ep_optimiser(
            self,
            optimiser: AbstractFactorOptimiser,
            paths: Optional[AbstractPaths] = None,
            ep_history: Optional = None,
    ) -> EPOptimiser:
        return EPOptimiser(
            self.graph,
            default_optimiser=optimiser,
            factor_optimisers={
                factor: factor.optimiser
                for factor in self.model_factors
                if factor.optimiser is not None
            },
            ep_history=ep_history,
            paths=paths
        )

    def optimise(
            self,
            optimiser: AbstractFactorOptimiser,
            paths: Optional[AbstractPaths] = None,
            ep_history: Optional = None,
            **kwargs
    ):
        """
        Use an EP Optimiser to optimise the graph associated with this collection
        of factors and create a Collection to represent the results.

        Parameters
        ----------
        paths
            Optionally define how data should be output. This paths
            object is copied to every optimiser.
        optimiser
            An optimiser that acts on graphs

        Returns
        -------
        ep_result: EPResult
            A collection of prior models
        """
        from autofit.graphical.declarative.result import EPResult
        opt = self._make_ep_optimiser(
            optimiser,
            paths=paths,
            ep_history=ep_history
        )
        updated_ep_mean_field = opt.run(
            self.mean_field_approximation(),
            **kwargs
        )

        return EPResult(
            ep_history=opt.ep_history,
            declarative_factor=self,
            updated_ep_mean_field=updated_ep_mean_field,
        )

    def visualize(
            self,
            paths: AbstractPaths,
            instance: ModelInstance,
            during_analysis: bool
    ):
        """
        Visualise the instances provided using each factor.

        Instances in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        instance
            A collection of instances, each corresponding to a factor
        during_analysis
            Is this visualisation during analysis?
        """
        for model_factor, instance in zip(
                self.model_factors,
                instance
        ):
            model_factor.visualize(
                paths,
                instance,
                during_analysis
            )

    @property
    def global_prior_model(self) -> Collection:
        """
        A collection of prior models, with one model for each factor.
        """
        return GlobalPriorModel(self)


class GlobalPriorModel(Collection):
    def __init__(
            self,
            factor: AbstractDeclarativeFactor
    ):
        """
        A global model comprising all factors which can be used to compare
        results between global optimisation and expectation propagation.

        Parameters
        ----------
        factor
            A factor comprising one or more factors, usually a graph
        """
        super().__init__([
            model_factor.prior_model
            for model_factor
            in factor.model_factors
        ])
        self.factor = factor

    @property
    def info(self) -> str:
        """
        A string describing the collection of factors in the graphical style
        """
        return self.factor.graph.info



import logging
from typing import Tuple

from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior.tuple_prior import TuplePrior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.collection import Collection
from .analysis import Analysis
from .indexed import IndexCollectionAnalysis
from ..paths.abstract import AbstractPaths


logger = logging.getLogger(__name__)


class FreeParameterAnalysis(IndexCollectionAnalysis):
    def __init__(self, *analyses: Analysis, free_parameters: Tuple[Prior, ...]):
        """
        A combined analysis with free parameters.

        All parameters for the model are shared across every analysis except
        for the free parameters which are allowed to vary for individual
        analyses.

        Parameters
        ----------
        analyses
            A list of analyses
        free_parameters
            A list of priors which are independent for each analysis
        """
        super().__init__(*analyses)
        self.free_parameters = [
            parameter for parameter in free_parameters if isinstance(parameter, Prior)
        ]
        # noinspection PyUnresolvedReferences
        self.free_parameters += [
            prior
            for parameter in free_parameters
            if isinstance(parameter, (AbstractPriorModel, TuplePrior))
            for prior in parameter.priors
        ]

    def modify_model(self, model: AbstractPriorModel) -> AbstractPriorModel:
        """
        Create prior models where free parameters are replaced with new
        priors. Return those prior models as a collection.

        The number of dimensions of the new prior model is the number of the
        old one plus the number of free parameters multiplied by the number
        of free parameters.

        Parameters
        ----------
        model
            The original model

        Returns
        -------
        A new model with all the same priors except for those associated
        with free parameters.
        """
        return Collection(
            [
                analysis.modify_model(
                    model.mapper_from_partial_prior_arguments(
                        {
                            free_parameter: free_parameter.new()
                            for free_parameter in self.free_parameters
                        }
                    )
                )
                for analysis in self.analyses
            ]
        )

    def modify_before_fit(self, paths: AbstractPaths, model: Collection):
        """
        Modify the analysis before fitting.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        model
            The model which is to be fitted.
        """
        return FreeParameterAnalysis(
            *(
                analysis.modify_before_fit(paths, model_)
                for analysis, model_ in zip(self.analyses, model)
            ),
            free_parameters=tuple(self.free_parameters),
        )

    def modify_after_fit(self, paths: AbstractPaths, model: Collection, result):
        """
        Modify the analysis after fitting.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        model
            The model which is to be fitted.
        result
            The result of the fit.
        """
        return FreeParameterAnalysis(
            *(
                analysis.modify_after_fit(paths, model, result)
                for analysis, model_ in zip(self.analyses, model)
            ),
            free_parameters=tuple(self.free_parameters),
        )



from functools import wraps


def cast_collection(named_tuple):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return list(map(lambda tup: named_tuple(*tup), func(*args, **kwargs)))

        return wrapper

    return decorator


class AttributeNameValue:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __iter__(self):
        return iter(self.tuple)

    @property
    def tuple(self):
        return self.name, self.value

    def __getitem__(self, item):
        return self.tuple[item]

    def __eq__(self, other):
        if isinstance(other, AttributeNameValue):
            return self.tuple == other.tuple
        if isinstance(other, tuple):
            return self.tuple == other
        return False

    def __hash__(self):
        return hash(self.tuple)

    def __str__(self):
        return "({}, {})".format(self.name, self.value)

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, str(self))


class PriorNameValue(AttributeNameValue):
    @property
    def prior(self):
        return self.value


class InstanceNameValue(AttributeNameValue):
    @property
    def instance(self):
        return self.value


class DeferredNameValue(AttributeNameValue):
    @property
    def deferred(self):
        return self.value



import copy
import itertools
from typing import Type, Union, Tuple

from autoconf.class_path import get_class
from .identifier import Identifier


class ModelObject:
    _ids = itertools.count()

    @classmethod
    def next_id(cls):
        return next(cls._ids)

    def __init__(
        self, id_=None, label=None,
    ):
        """
        A generic object in AutoFit

        Parameters
        ----------
        id_
            A unique integer identifier. This is used to hash and order priors.
        label
            A label which can optionally be set for visualising this object in a
            graph.
        """
        self.id = id_ or self.next_id()
        self._label = label

    def replacing_for_path(self, path: Tuple[str, ...], value) -> "ModelObject":
        """
        Create a new model replacing the value for a given path with a new value

        Parameters
        ----------
        path
            A path indicating the sequence of names used to address an object
        value
            A value that should replace the object at the given path

        Returns
        -------
        A copy of this with an updated value
        """
        new = copy.deepcopy(self)
        obj = new
        for key in path[:-1]:
            obj = getattr(obj, key)

        setattr(obj, path[-1], value)
        return new

    def has(self, cls: Union[Type, Tuple[Type, ...]]) -> bool:
        """
        Does this instance have an attribute which is of type cls?
        """
        for value in self.__dict__.values():
            if isinstance(value, cls):
                return True
        return False

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def component_number(self):
        return self.id

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    @property
    def identifier(self):
        return str(Identifier(self))

    @staticmethod
    def from_dict(d):
        """
        Recursively parse a dictionary returning the model, collection or
        instance that is represents.

        Parameters
        ----------
        d
            A dictionary representation of some object

        Returns
        -------
        An instance
        """
        from autofit.mapper.prior_model.abstract import AbstractPriorModel
        from autofit.mapper.prior_model.collection import Collection
        from autofit.mapper.prior_model.prior_model import Model
        from autofit.mapper.prior.abstract import Prior
        from autofit.mapper.prior.tuple_prior import TuplePrior

        if not isinstance(d, dict):
            return d

        type_ = d["type"]

        if type_ == "model":
            instance = Model(get_class(d.pop("class_path")))
        elif type_ == "collection":
            instance = Collection()
        elif type_ == "tuple_prior":
            instance = TuplePrior()
        elif type_ == "dict":
            return {key: ModelObject.from_dict(value) for key, value in d.items()}
        elif type_ == "instance":
            d.pop("type")
            cls = get_class(d.pop("class_path"))
            return cls(
                **{key: ModelObject.from_dict(value) for key, value in d.items()}
            )
        else:
            try:
                return Prior.from_dict(d)
            except KeyError:
                cls = get_class(type_)
                instance = object.__new__(cls)

        d.pop("type")

        for key, value in d.items():
            setattr(instance, key, AbstractPriorModel.from_dict(value))
        return instance

    def dict(self) -> dict:
        """
        A dictionary representation of this object
        """
        from autofit.mapper.prior_model.abstract import AbstractPriorModel
        from autofit.mapper.prior_model.collection import Collection
        from autofit.mapper.prior_model.prior_model import Model
        from autofit.mapper.prior.tuple_prior import TuplePrior

        if isinstance(self, Collection):
            type_ = "collection"
        elif isinstance(self, AbstractPriorModel) and self.prior_count == 0:
            type_ = "instance"
        elif isinstance(self, Model):
            type_ = "model"
        elif isinstance(self, TuplePrior):
            type_ = "tuple_prior"
        else:
            raise AssertionError(
                f"{self.__class__.__name__} cannot be serialised to dict"
            )

        dict_ = {"type": type_}

        for key, value in self._dict.items():
            try:
                if not isinstance(value, ModelObject):
                    value = AbstractPriorModel.from_instance(value)
                value = value.dict()
            except AttributeError:
                pass
            except TypeError:
                pass
            dict_[key] = value
        return dict_

    @property
    def _dict(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ("component_number", "item_number", "id", "cls")
            and not key.startswith("_")
        }



from abc import abstractmethod
from typing import List, Optional
import warnings

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.pdf import SamplesPDF
from autofit.non_linear.samples.samples import Sample
from autofit.non_linear.samples.stored import SamplesStored

from autofit import exc

class SamplesNest(SamplesPDF):

    def __init__(
            self,
            model: AbstractPriorModel,
            sample_list: List[Sample],
            number_live_points: Optional[int] = None,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
            results_internal: Optional = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results_internal of
        a `NonLinearSearch` (e.g. as files on your hard-disk) and Python.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        This class stores the samples of nested sampler model-fit (e.g. `dynesty`, `UltraNest`). To use a library's
        in-built visualization tools results are optionally stored in their native internal format using the
        `results_internal` attribute.

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        sample_list
            The list of `Samples` which contains the paramoeters, likelihood, weights, etc. of every sample taken
            by the non-linear search.
        number_live_points
            The number of live points used by the nested sampler.
        unconverged_sample_size
            If the samples are for a search that is yet to convergence, a reduced set of samples are used to provide
            a rough estimate of the parameters. The number of samples is set by this parameter.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        results_internal
            The nested sampler's results in their native internal format for interfacing its visualization library.
        """

        self._number_live_points = number_live_points

        super().__init__(
            model=model,
            sample_list=sample_list,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
            results_internal=results_internal
        )

    def __add__(
            self,
            other: "SamplesNest"
    ) -> "SamplesNest":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        For UltraNest samples there are no tools for combining results in their native format, therefore these
        `results_internal` are set to None and support for visualization is disabled.

        Parameters
        ----------
        other
            Another Samples class

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        warnings.warn(
            f"Addition of {self.__class__.__name__} cannot retain results in native format. "
            "Visualization of summed samples diabled.",
            exc.SamplesWarning
        )

        return self.__class__(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
            number_live_points=self._number_live_points,
            unconverged_sample_size=self.unconverged_sample_size,
            time=self.time,
            results_internal=None
        )

    @property
    def number_live_points(self):
        return self._number_live_points

    @property
    @abstractmethod
    def log_evidence(self):
        pass

    @property
    @abstractmethod
    def total_samples(self):
        pass

    @property
    def info_json(self):
        return {
            "log_evidence": self.log_evidence,
            "total_samples": self.total_samples,
            "unconverged_sample_size": self.unconverged_sample_size,
            "time": self.time,
            "number_live_points": self.number_live_points
        }

    @property
    def total_accepted_samples(self) -> int:
        """
        The total number of accepted samples performed by the nested sampler.
        """
        return len(self.log_likelihood_list)

    @property
    def acceptance_ratio(self) -> float:
        """
        The ratio of accepted samples to total samples.
        """
        return self.total_accepted_samples / self.total_samples

    def samples_within_parameter_range(
            self,
            parameter_index: int,
            parameter_range: [float, float]
    ) -> "SamplesStored":
        """
        Returns a new set of Samples where all points without parameter values inside a specified range removed.

        For example, if our `Samples` object was for a model with 4 parameters are consistent of the following 3 sets
        of parameters:

        [[1.0, 2.0, 3.0, 4.0]]
        [[100.0, 2.0, 3.0, 4.0]]
        [[1.0, 2.0, 3.0, 4.0]]

        This function for `parameter_index=0` and `parameter_range=[0.0, 99.0]` would remove the second sample because
        the value 100.0 is outside the range 0.0 -> 99.0.

        Parameters
        ----------
        parameter_index
            The 1D index of the parameter (in the model's vector representation) whose values lie between the parameter
            range if a sample is kept.
        parameter_range
            The minimum and maximum values of the range of parameter values this parameter must lie between for it
            to be kept.
        """

        parameter_list = []
        log_likelihood_list = []
        log_prior_list = []
        weight_list = []

        for sample in self.sample_list:

            parameters = sample.parameter_lists_for_model(model=self.model)

            if (parameters[parameter_index] > parameter_range[0]) and (
                    parameters[parameter_index] < parameter_range[1]):

                parameter_list.append(parameters)
                log_likelihood_list.append(sample.log_likelihood)
                log_prior_list.append(sample.log_prior)
                weight_list.append(sample.weight)

        sample_list = Sample.from_lists(
            model=self.model,
            parameter_lists=parameter_list,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        return SamplesStored(
            model=self.model,
            sample_list=sample_list,
            unconverged_sample_size=self.unconverged_sample_size,
        )



import shutil
from typing import Optional

from autofit.database.sqlalchemy_ import sa
from .abstract import AbstractPaths
from ...database.model import Fit


class DatabasePaths(AbstractPaths):
    def __init__(
        self,
        session,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths=True,
        parent=None,
        save_all_samples=False,
        unique_tag: Optional["str"] = None,
    ):
        super().__init__(
            name=name,
            path_prefix=path_prefix,
            is_identifier_in_paths=is_identifier_in_paths,
            parent=parent,
        )
        self.session = session
        self._fit = None
        self.save_all_samples = save_all_samples
        self.unique_tag = unique_tag

    parent: "DatabasePaths"

    @AbstractPaths.parent.setter
    def parent(self, parent: "DatabasePaths"):
        """
        The search performed before this search. For example, a search
        that is then compared to searches during a grid search.

        For database paths the parent must also be database paths.
        """
        if not (parent is None or isinstance(parent, DatabasePaths)):
            raise TypeError(
                "The parent of search that uses the database must also use the database"
            )
        self._parent = parent

    def save_named_instance(self, name: str, instance):
        """
        Save an instance, such as that at a given sigma
        """
        self.fit.named_instances[name] = instance

    @property
    def is_grid_search(self) -> bool:
        return self.fit.is_grid_search

    def create_child(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        is_identifier_in_paths: Optional[bool] = None,
        identifier: Optional[str] = None,
    ) -> "DatabasePaths":
        """
        Create a paths object which is the child of some parent
        paths object. This is done during a GridSearch so that
        results can be stored in the correct directory. It also
        allows database fit objects to be related correctly.

        If no instance is set the prior median model is used
        to ensure that the parent object is queryable.

        Parameters
        ----------
        identifier
        name
        path_prefix
        is_identifier_in_paths
            If False then this path's identifier will not be
            added to its output path.

        Returns
        -------
        A new paths object
        """
        self.fit.is_grid_search = True
        if self.fit.instance is None:
            self.fit.instance = self.model.instance_from_prior_medians()
        child = type(self)(
            session=self.session,
            name=name or self.name,
            path_prefix=path_prefix or self.path_prefix,
            is_identifier_in_paths=(
                is_identifier_in_paths
                if is_identifier_in_paths is not None
                else self.is_identifier_in_paths
            ),
            parent=self,
        )
        child.model = self.model
        child.search = self.search
        child._identifier = identifier
        return child

    def zip_remove(self):
        """
        Remove files from both the symlinked folder and the output directory
        """
        self.session.commit()

        if self.remove_files:
            shutil.rmtree(self.output_path, ignore_errors=True)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d["session"]
        return d

    def save_object(self, name: str, obj: object):
        self.fit[name] = obj

    def load_object(self, name: str):
        return self.fit[name]

    def remove_object(self, name: str):
        del self.fit[name]

    def is_object(self, name: str) -> bool:
        return name in self.fit

    @property
    def fit(self) -> Fit:
        if self._fit is None:
            try:
                self._fit = (
                    self.session.query(Fit).filter(Fit.id == self.identifier).one()
                )
            except sa.orm.exc.NoResultFound:
                self._fit = Fit(
                    id=self.identifier,
                    is_complete=False,
                    unique_tag=self.unique_tag,
                    path_prefix=self.path_prefix,
                    name=self.name,
                )
                self.session.add(self._fit)

        if self.parent is not None:
            self._fit.parent = self.parent.fit
        return self._fit

    @property
    def is_complete(self) -> bool:
        return self.fit.is_complete

    def completed(self):
        self.fit.is_complete = True

    def save_summary(self, samples, log_likelihood_function_time):
        self.fit.instance = samples.max_log_likelihood()
        self.fit.max_log_likelihood = samples.max_log_likelihood_sample.log_likelihood
        super().save_summary(samples, log_likelihood_function_time)

    def save_samples(self, samples):
        if not self.save_all_samples:
            samples = samples.minimise()

        self.fit.samples = samples

    def samples_to_csv(self, samples):
        """
        Save the final-result samples associated with the phase as a pickle
        """
        pass

    def _load_samples(self):
        samples = self.fit.samples
        samples.model = self.model
        return samples

    def load_samples(self):
        return self._load_samples().sample_list

    def load_samples_info(self):
        return self._load_samples().info_json

    def save_all(self, info, *_, **kwargs):
        self.save_identifier()
        self.fit.info = info
        self.fit.model = self.model

        self.save_object("search", self.search)

        self.session.commit()



from collections.abc import Iterable

from autofit.mapper.model import ModelInstance, assert_not_frozen
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.abstract import check_assertions


class Collection(AbstractPriorModel):
    def name_for_prior(self, prior: Prior) -> str:
        """
        Construct a name for the prior. This is the path taken
        to get to the prior.

        Parameters
        ----------
        prior

        Returns
        -------
        A string of object names joined by underscores
        """
        for name, prior_model in self.prior_model_tuples:
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(name, prior_name)
        for name, direct_prior in self.direct_prior_tuples:
            if prior == direct_prior:
                return name

    def __contains__(self, item):
        return item in self._dict or item in self._dict.values()

    def __getitem__(self, item):
        if item in self._dict:
            return self._dict[item]
        return self.values[item]

    def __len__(self):
        return len(self.values)

    def __str__(self):
        return "\n".join(
            f"{key} = {value}"
            for key, value
            in self.items()
        )

    def __hash__(self):
        return self.id

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    @property
    def values(self):
        return list(self._dict.values())

    def items(self):
        return self._dict.items()

    def with_prefix(
            self,
            prefix: str
    ):
        """
        Filter members of the collection, only returning those that start
        with a given prefix as a new collection.
        """
        return Collection({
            key: value
            for key, value
            in self.items()
            if key.startswith(
                prefix
            )
        })

    def as_model(self):
        return Collection({
            key: value.as_model()
            if isinstance(value, AbstractPriorModel)
            else value
            for key, value in self.dict().items()
        })

    def __init__(self, *arguments, **kwargs):
        """
        The object multiple Python classes are input into to create model-components, which has free parameters that
        are fitted by a non-linear search.

        Multiple Python classes can be input into a `Collection` in order to compose high dimensional models made of
        multiple model-components.

        The ``Collection`` object is highly flexible, and can create models from many input Python data structures
        (e.g. a list of classes, dictionary of classes, hierarchy of classes).

        For a complete description of the model composition API, see the **PyAutoFit** model API cookbooks:

        https://pyautofit.readthedocs.io/en/latest/cookbooks/cookbook_1_basics.html

        The Python class input into a ``Model`` to create a model component is written using the following format:

        - The name of the class is the name of the model component (e.g. ``Gaussian``).
        - The input arguments of the constructor are the parameters of the mode (e.g. ``centre``, ``normalization`` and ``sigma``).
        - The default values of the input arguments tell PyAutoFit whether a parameter is a single-valued float or a
        multi-valued tuple.

        [Rich document more clearly]

        A prior model used to represent a list of prior models for convenience.

        Arguments are flexibly converted into a collection.

        Parameters
        ----------
        arguments
            Classes, prior models, instances or priors

        Examples
        --------

        class Gaussian:

            def __init__(
                self,
                centre=0.0,        # <- PyAutoFit recognises these
                normalization=0.1, # <- constructor arguments are
                sigma=0.01,        # <- the Gaussian's parameters.
            ):
                self.centre = centre
                self.normalization = normalization
                self.sigma = sigma

        model = af.Collection(gaussian_0=Gaussian, gaussian_1=Gaussian)
        """
        super().__init__()
        self.item_number = 0
        arguments = list(arguments)
        if len(arguments) == 0:
            self.add_dict_items(kwargs)
        elif len(arguments) == 1:
            arguments = arguments[0]

            if isinstance(arguments, dict):
                self.add_dict_items(arguments)
            elif isinstance(arguments, Iterable):
                for argument in arguments:
                    self.append(argument)
            else:
                self.append(arguments)
        else:
            self.__init__(arguments)

    @assert_not_frozen
    def add_dict_items(self, item_dict):
        for key, value in item_dict.items():
            if isinstance(key, tuple):
                key = ".".join(key)
            setattr(self, key, AbstractPriorModel.from_object(value))

    def __eq__(self, other):
        if other is None:
            return False
        if len(self) != len(other):
            return False
        for i, item in enumerate(self):
            if item != other[i]:
                return False
        return True

    @assert_not_frozen
    def append(self, item):
        setattr(self, str(self.item_number), AbstractPriorModel.from_object(item))
        self.item_number += 1

    @assert_not_frozen
    def __setitem__(self, key, value):
        obj = AbstractPriorModel.from_object(value)
        try:
            obj.id = getattr(self, str(key)).id
        except AttributeError:
            pass
        setattr(self, str(key), obj)

    @assert_not_frozen
    def __setattr__(self, key, value):
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            try:
                super().__setattr__(key, AbstractPriorModel.from_object(value))
            except AttributeError:
                pass

    def remove(self, item):
        for key, value in self.__dict__.copy().items():
            if value == item:
                del self.__dict__[key]

    @check_assertions
    def _instance_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        model_instances: [object]
            A list of instances constructed from the list of prior models.
        """
        result = ModelInstance()
        for key, value in self.__dict__.items():
            if key.startswith("_"):
                continue
            if isinstance(value, AbstractPriorModel):
                value = value.instance_for_arguments(arguments)
            elif isinstance(value, Prior):
                value = arguments[value]
            setattr(result, key, value)
        return result

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Create a new collection, updating its priors according to the argument
        dictionary.

        Parameters
        ----------
        arguments
            A dictionary of arguments

        Returns
        -------
        A new collection
        """
        collection = Collection()

        for key, value in self.items():
            if key in (
                    "component_number",
                    "item_number",
                    "id"
            ) or key.startswith(
                "_"
            ):
                continue

            if isinstance(value, AbstractPriorModel):
                collection[key] = value.gaussian_prior_model_for_arguments(
                    arguments
                )
            if isinstance(value, Prior):
                collection[key] = arguments[value]

        return collection

    @property
    def prior_class_dict(self):
        return {
            **{
                prior: cls
                for prior_model in self.direct_prior_model_tuples
                for prior, cls in prior_model[1].prior_class_dict.items()
            },
            **{prior: ModelInstance for _, prior in self.direct_prior_tuples},
        }



from autofit.graphical.factor_graphs.factor import FactorKW
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.collection import Collection
from autofit.non_linear.analysis import Analysis
from autofit.tools.namer import namer


class PriorFactor(FactorKW, Analysis):
    def __init__(self, prior: Prior):
        """
        A factor that wraps a prior such that is can be optimised
        by classic AutoFit optimisers.

        To do this it implements a prior_model and analysis.

        Parameters
        ----------
        prior
            A message/prior, usually of another factor. Prior factors
            are generated programmatically.
        """
        # TODO: Consider analytical solution rather than implementing optimisation
        super().__init__(prior.factor, x=prior, name=namer(self.__class__.__name__))
        self.prior = prior
        self.label = f"PriorFactor({prior.label})"

    @property
    def _unique_representation(self):
        return (
            self.prior,
            self.arg_names,
            self.args,
            self.deterministic_variables,
        )

    @property
    def prior_model(self) -> Collection:
        """
        A trivial prior model to conform to the expected interface.
        """
        return Collection(self.prior)

    @property
    def analysis(self) -> "PriorFactor":
        """
        This is the analysis class for a PriorFactor
        """
        return self

    def log_likelihood_function(self, instance) -> float:
        """
        Compute the likelihood.

        The instance is a collection with a single argument expressing a
        possible value for this prior. The likelihood is computed by simply
        evaluating the prior's PDF for the given value.
        """
        return self.prior.factor(instance[0])

    @property
    def variable(self):
        return list(self.variables)[0]



import inspect
import logging
from abc import ABC
from copy import copy

import numpy as np

from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.mapper.prior_model.abstract import AbstractPriorModel

logger = logging.getLogger(
    __name__
)


def retrieve_name(var):
    first_name = None
    frame = inspect.currentframe()
    while frame is not None:
        for name, value in frame.f_locals.items():
            if var is value:
                first_name = name
        frame = frame.f_back

    return first_name


class CompoundPrior(
    AbstractPriorModel,
    ArithmeticMixin,
    ABC
):
    cls = float

    def __init__(self, left, right):
        """
        Comprises objects that are to undergo some arithmetic
        operation after realisation.

        Parameters
        ----------
        left
            A prior, promise or float
        right
            A prior, promise or float
        """
        super().__init__()

        self._left_name = retrieve_name(left) or "left"
        self._right_name = retrieve_name(right) or "right"

        if self._left_name == "left":
            self._left_name = "left_"

        if self._right_name == "right":
            self._right_name = "right_"

        self._left = None
        self._right = None

        self.left = left
        self.right = right

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @left.setter
    def left(self, left):
        self._left = left
        setattr(self, self._left_name, left)

    @right.setter
    def right(self, right):
        self._right = right
        setattr(self, self._right_name, right)

    def gaussian_prior_model_for_arguments(self, arguments):
        new = copy(self)
        try:
            new.left = new.left.gaussian_prior_model_for_arguments(
                arguments
            )
        except AttributeError:
            pass
        try:
            new.right = new.right.gaussian_prior_model_for_arguments(
                arguments
            )
        except AttributeError:
            pass
        return new

    def left_for_arguments(
            self,
            arguments: dict
    ):
        """
        Instantiate the left object.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to values

        Returns
        -------
        A value for the left object
        """
        try:
            return self._left.instance_for_arguments(arguments, )
        except AttributeError:
            return self._left

    def right_for_arguments(
            self,
            arguments: dict
    ):
        """
        Instantiate the right object.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to values

        Returns
        -------
        A value for the right object
        """
        try:
            return self._right.instance_for_arguments(arguments, )
        except AttributeError:
            return self._right


class SumPrior(CompoundPrior):
    """
    The sum of two objects, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) + self.right_for_arguments(
            arguments
        )

    def __str__(self):
        return f"{self._left} + {self._right}"


class MultiplePrior(CompoundPrior):
    """
    The multiple of two objects, computed after realisation.
    """

    def __str__(self):
        return f"{self._left} * {self._right}"

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) * self.right_for_arguments(
            arguments
        )


class DivisionPrior(CompoundPrior):
    """
    One object divided by another, computed after realisation
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) / self.right_for_arguments(
            arguments
        )


class FloorDivPrior(CompoundPrior):
    """
    One object divided by another and floored, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) // self.right_for_arguments(
            arguments
        )


class ModPrior(CompoundPrior):
    """
    The modulus of a pair of objects, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) % self.right_for_arguments(
            arguments
        )


class PowerPrior(CompoundPrior):
    """
    One object to the power of another, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return self.left_for_arguments(
            arguments
        ) ** self.right_for_arguments(
            arguments
        )


class ModifiedPrior(
    AbstractPriorModel,
    ABC,
    ArithmeticMixin
):
    def __init__(self, prior):
        super().__init__()
        self._prior_name = retrieve_name(prior)

        if self._prior_name == "prior":
            self._prior_name = "prior_"

        self.prior = prior

    def __add__(self, other):
        return ArithmeticMixin.__add__(self, other)

    @property
    def cls(self):
        return self.prior.cls

    @property
    def prior(self):
        return getattr(self, self._prior_name)

    @prior.setter
    def prior(self, prior):
        setattr(self, self._prior_name, prior)

    def gaussian_prior_model_for_arguments(self, arguments):
        new = copy(self)
        try:
            new.prior = new.prior.gaussian_prior_model_for_arguments(
                arguments
            )
        except AttributeError:
            pass
        return new


class NegativePrior(ModifiedPrior):
    """
    The negation of an object, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return -self.prior.instance_for_arguments(arguments, )


class AbsolutePrior(ModifiedPrior):
    """
    The absolute value of an object, computed after realisation.
    """

    def _instance_for_arguments(self, arguments):
        return abs(self.prior.instance_for_arguments(arguments, ))


class Log(ModifiedPrior):
    """
    The natural logarithm of an object, computed after realisation.
    """

    def _instance_for_arguments(
            self,
            arguments
    ):
        return np.log(self.prior.instance_for_arguments(arguments, ))


class Log10(ModifiedPrior):
    """
    The base10 logarithm of an object, computed after realisation.
    """

    def _instance_for_arguments(
            self,
            arguments
    ):
        return np.log10(self.prior.instance_for_arguments(arguments, ))



"""
This module provides a wrapper over the scipy.optimize.linesearch module

To work with the Factor and FactorJacobian interface defined in 
autofit.graphical

Note that this interface assumes that we're performing a maximisation. 
In scipy the interface is defined for minimisations.
"""
import warnings
from typing import Optional, Dict, Tuple

import numpy as np
from scipy.optimize import linesearch

from autoconf import cached_property
from autofit.graphical.factor_graphs.abstract import (
    FactorValue,
    FactorInterface,
    FactorGradientInterface,
)
from autofit.graphical.utils import FlattenArrays
from autofit.mapper.variable_operator import (
    VariableData,
    VariableLinearOperator,
    MergedVariableOperator,
)


class FlattenedState:
    def __init__(self, state, param_shapes):
        self.state = state
        self.param_shapes = param_shapes

    @classmethod
    def from_state(cls, state):
        param_shapes = FlattenArrays.from_arrays(state.parameters)
        return cls(state, param_shapes)

    def make_state(self, x):
        return self.state.update(parameters=self.param_shapes.unflatten(x))

    def __call__(self, x):
        new_state = self.make_state(x)
        return new_state.value

    def func_gradient(self, x):
        new_state = self.make_state(x)
        val, grad = new_state.value_gradient
        return val, self.param_shapes.flatten(grad)

    def _func(self, x):
        return -self(x)

    def _func_gradient(self, x):
        v, g = self.func_gradient(x)
        return -v, -g

    @property
    def parameters(self):
        return self.param_shapes.flatten(self.state.parameters)


class OptimisationState:
    def __init__(
        self,
        factor: FactorInterface,
        factor_gradient: FactorGradientInterface,
        parameters: VariableData,
        hessian: Optional[VariableLinearOperator] = None,
        det_hessian: Optional[VariableLinearOperator] = None,
        value: Optional[FactorValue] = None,
        gradient: Optional[VariableData] = None,
        search_direction: Optional[VariableData] = None,
        f_count: int = 0,
        g_count: int = 0,
        args=(),
        next_states: Optional[Dict[float, "OptimisationState"]] = None,
        lower_limit=None,
        upper_limit=None,
    ):
        self.factor = factor
        self.factor_gradient = factor_gradient

        self._parameters = None
        self.parameters = parameters
        self.hessian = hessian
        self.det_hessian = det_hessian
        self.f_count = np.asanyarray(f_count)
        self.g_count = np.asanyarray(g_count)
        self.args = args
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

        self.next_states = next_states or {}

        if not self.valid:
            value = -FactorValue(np.inf)
            gradient = self.parameters.full_like(np.inf)

        self._value = value
        self._gradient = gradient

        if search_direction is not None:
            self.search_direction = search_direction

    @property
    def valid(self):
        if self.lower_limit and (self.parameters < self.lower_limit).any():
            return False

        if self.upper_limit and (self.parameters > self.upper_limit).any():
            return False

        return True

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        # This forces recalculation of the value and gradient as needed
        self._value = None
        self._gradient = None
        self._parameters = parameters

    @property
    def value(self):
        if self._value is None:
            self.f_count += 1
            self._value = self.factor(self.parameters, *self.args)

        return self._value

    @property
    def gradient(self):
        if self._gradient is None:
            self._gradient = self.value_gradient[1]
        return self._gradient

    @cached_property
    def value_gradient(self):
        self.g_count += 1
        self._value, self._gradient = val = self.factor_gradient(
            self.parameters, *self.args
        )
        return val

    def to_dict(self):
        # don't return value, gradient or search direction as may change
        return {
            "factor": self.factor,
            "factor_gradient": self.factor_gradient,
            "parameters": self.parameters,
            "hessian": self.hessian,
            "det_hessian": self.det_hessian,
            "f_count": self.f_count,
            "g_count": self.g_count,
            "args": self.args,
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit,
        }

    def copy(self):
        return type(self)(**self.to_dict())

    def update(self, **kwargs):
        return type(self)(**{**self.to_dict(), **kwargs})

    def __repr__(self):
        vals = self.to_dict()
        if vals.keys():
            m = max(map(len, list(vals.keys()))) + 1
            attrs = "\n".join(
                [k.rjust(m) + " = " + repr(v) + "," for k, v in vals.items()]
            )
            return self.__class__.__name__ + f"(\n{attrs}\n)"
        else:
            return self.__class__.__name__ + "()"

    def _next_state(self, stepsize):
        next_params = VariableData.add(
            self.parameters, VariableData.mul(self.search_direction, stepsize)
        )
        # memoize stepsizes
        self.next_states[stepsize] = next_state = self.update(parameters=next_params)
        return next_state

    def step(self, stepsize):

        if not stepsize:
            return self

        stepsize = float(stepsize)
        # memoize stepsizes
        next_state = self.next_states.get(stepsize) or self._next_state(stepsize)

        return next_state

    def phi(self, s):
        next_state = self.step(s)
        return -next_state.value

    def derphi(self, s):
        next_state = self.step(s)
        return self.calc_derphi(next_state.gradient)

    def calc_derphi(self, gradient):
        return -VariableData.dot(self.search_direction, gradient)

    @property
    def all_parameters(self):
        return self.parameters.merge(self.value.deterministic_values)

    @property
    def full_hessian(self):
        if self.det_hessian:
            return MergedVariableOperator(self.hessian, self.det_hessian)

        return self.hessian

    def hessian_blocks(self):
        blocks = self.hessian.blocks()
        if self.det_hessian:
            blocks.update(self.det_hessian.blocks())

        return blocks

    def inv_hessian_blocks(self):
        blocks = self.hessian.inv().blocks()
        if self.det_hessian:
            blocks.update(self.det_hessian.inv().blocks())

        return blocks

    def hessian_diagonal(self):
        diagonal = self.hessian.diagonal()
        if self.det_hessian:
            diagonal.update(self.det_hessian.diagonal())
        return diagonal

    def flatten(self):
        return FlattenedState.from_state(self)


def line_search_wolfe1(
    state: OptimisationState,
    old_state: Optional[OptimisationState] = None,
    c1=1e-4,
    c2=0.9,
    amax=50,
    amin=1e-8,
    xtol=1e-14,
    extra_condition=None,
    **kwargs,
) -> Tuple[Optional[float], OptimisationState]:
    """
    As `scalar_search_wolfe1` but do a line search to direction `pk`
    Parameters
    ----------
    f : callable
        Function `f(x)`
    fprime : callable
        Gradient of `f`
    xk : array_like
        Current point
    pk : array_like
        Search direction
    gfk : array_like, optional
        Gradient of `f` at point `xk`
    old_fval : float, optional
        Value of `f` at point `xk`
    old_old_fval : float, optional
        Value of `f` at point preceding `xk`
    The rest of the parameters are the same as for `scalar_search_wolfe1`.
    Returns
    -------
    stp, f_count, g_count, fval, old_fval
        As in `line_search_wolfe1`
    gval : array
        Gradient of `f` at the final point
    """
    derphi0 = state.derphi(0)
    old_fval = state.value
    stepsize, _, _ = linesearch.scalar_search_wolfe1(
        state.phi,
        state.derphi,
        -old_fval,  # we are actually performing maximisation
        old_state and -old_state.value,
        derphi0,
        c1=c1,
        c2=c2,
        amax=amax,
        amin=amin,
        xtol=xtol,
    )
    next_state = state.step(stepsize)

    if stepsize is not None and extra_condition is not None:
        if not extra_condition(stepsize, next_state):
            stepsize = None

    return stepsize, next_state


def line_search_wolfe2(
    state: OptimisationState,
    old_state: Optional[OptimisationState] = None,
    c1=1e-4,
    c2=0.9,
    amax=None,
    extra_condition=None,
    maxiter=10,
    **kwargs,
) -> Tuple[Optional[float], OptimisationState]:
    """
    As `scalar_search_wolfe1` but do a line search to direction `pk`
    Parameters
    ----------
    f : callable
        Function `f(x)`
    fprime : callable
        Gradient of `f`
    xk : array_like
        Current point
    pk : array_like
        Search direction
    gk : array_like, optional
        Gradient of `f` at point `xk`
    old_fval : float, optional
        Value of `f` at point `xk`
    old_old_fval : float, optional
        Value of `f` at point preceding `xk`
    The rest of the parameters are the same as for `scalar_search_wolfe1`.
    Returns
    -------
    stp, f_count, g_count, fval, old_fval
        As in `line_search_wolfe1`
    gval : array
        Gradient of `f` at the final point
    """
    derphi0 = state.derphi(0)
    old_fval = state.value
    stepsize, _, _, _ = linesearch.scalar_search_wolfe2(
        state.phi,
        state.derphi,
        -old_fval,  # we are actually performing maximisation
        old_state and -old_state.value,
        derphi0,
        c1=c1,
        c2=c2,
        amax=amax,
        maxiter=maxiter,
    )

    next_state = state.step(stepsize)
    if stepsize is not None and extra_condition is not None:
        if not extra_condition(stepsize, next_state):
            stepsize = None

    return stepsize, next_state


def line_search(
    state: OptimisationState, old_state: Optional[FactorValue] = None, **kwargs
) -> Tuple[Optional[float], OptimisationState]:
    stepsize, next_state = line_search_wolfe1(state, old_state, **kwargs)

    if stepsize is None:
        # line search failed: try different one.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", linesearch.LineSearchWarning)
            stepsize, next_state = line_search_wolfe2(state, old_state, **kwargs)

    # if stepsize is None:
    #     raise _LineSearchError()

    return stepsize, next_state



import logging
from typing import Dict, List, Generator

from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.mean_field import Status
from autofit.graphical.utils import StatusFlag, LogWarnings
from autofit.mapper.variable import Plate
from autofit.tools.util import IntervalCounter
from .optimiser import EPOptimiser

logger = logging.getLogger(__name__)


class StochasticEPOptimiser(EPOptimiser):
    def factor_step(self, factor, subset_approx, optimiser):
        factor_logger = logging.getLogger(factor.name)
        factor_logger.debug("Optimising...")

        subset_factor = subset_approx._factor_subset_factor[factor]
        try:
            with LogWarnings(
                logger=factor_logger.debug, action="always"
            ) as caught_warnings:
                factor_approx = subset_approx.factor_approximation(factor)
                new_model_dist, status = optimiser.optimise(factor_approx)
                subset_approx, status = self.updater.update_model_approx(
                    new_model_dist, factor_approx, subset_approx, status
                )

            messages = status.messages + tuple(caught_warnings.messages)
            status = Status(status.success, messages, status.flag)
        except (ValueError, ArithmeticError, RuntimeError) as e:
            logger.exception(e)
            status = Status(
                False,
                status.messages + (f"Factor: {factor} experienced error {e}",),
                StatusFlag.FAILURE,
            )

        factor_logger.debug(status)
        return subset_approx, status

    def run(
        self,
        model_approx: EPMeanField,
        batches: Generator[Dict[Plate, List[int]], None, None],
        log_interval: int = 10,
        visualise_interval: int = 100,
        output_interval: int = 10,
        inplace=False,
    ) -> EPMeanField:
        """
        Run the optimisation on an approximation of the model.

        Parameters
        ----------
        model_approx
            A collection of messages describing priors on the model's variables.
        max_steps
            The maximum number of steps prior to termination. Termination may also
            occur when difference in log evidence or KL Divergence drop below a given
            threshold for two consecutive optimisations of a given factor.
        log_interval
            How steps should we wait before logging information?
        visualise_interval
            How steps should we wait before visualising information?
            This includes plots of KL Divergence and Evidence.
        output_interval
            How steps should we wait before outputting information?
            This includes the model.results file which describes the current mean values
            of each message.

        Returns
        -------
        An updated approximation of the model
        """
        should_log = IntervalCounter(log_interval)
        should_visualise = IntervalCounter(visualise_interval)
        should_output = IntervalCounter(output_interval)

        for batch in batches:
            subset_approx = model_approx[batch]

            for factor, optimiser in self.factor_optimisers.items():
                subset_approx, status = self.factor_step(
                    factor, subset_approx, optimiser
                )

                if status and should_log():
                    self._log_factor(factor)

                if self.ep_history(factor, subset_approx, status):
                    logger.info("Terminating optimisation")
                    break  # callback controls convergence

            else:  # If no break do next iteration

                if inplace:
                    model_approx[batch] = subset_approx
                else:
                    model_approx = model_approx.merge(batch, subset_approx)

                if self.ep_history(model_approx.factor_graph, model_approx):
                    logger.info("Terminating optimisation")
                    break  # callback controls convergence

                if should_visualise():
                    self.visualiser()
                if should_output():
                    self._output_results(model_approx)
                continue
            break  # stop iterations

        self.visualiser()
        self._output_results(model_approx)

        return model_approx



import os
from os import path
from typing import Optional

import emcee
import numpy as np

from autoconf import conf
from autofit.database.sqlalchemy_ import sa
from autofit.mapper.model_mapper import ModelMapper
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.mcmc.abstract_mcmc import AbstractMCMC
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.mcmc.emcee.samples import SamplesEmcee
from autofit.plot import EmceePlotter
from autofit.plot.output import Output


class Emcee(AbstractMCMC):
    __identifier_fields__ = ("nwalkers",)

    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        prior_passer: Optional[PriorPasser] = None,
        initializer: Optional[Initializer] = None,
        auto_correlations_settings=AutoCorrelationsSettings(),
        iterations_per_update: int = None,
        number_of_cores: int = None,
        session: Optional[sa.orm.Session] = None,
        **kwargs
    ):
        """
        An Emcee non-linear search.

        For a full description of Emcee, checkout its Github and readthedocs webpages:

        https://github.com/dfm/emcee

        https://emcee.readthedocs.io/en/stable/

        If you use `Emcee` as part of a published work, please cite the package following the instructions under the
        *Attribution* section of the GitHub page.

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        auto_correlations_settings
            Customizes and performs auto correlation calculations performed during and after the search.
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            auto_correlations_settings=auto_correlations_settings,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating Emcee Search")

    class Fitness(AbstractMCMC.Fitness):
        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space. `Emcee`
            uses the log posterior.
            """
            return self.log_posterior_from(parameter_list=parameter_list)

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using Emcee and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """
        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        pool = self.make_sneaky_pool(fitness_function)

        emcee_sampler = emcee.EnsembleSampler(
            nwalkers=self.config_dict_search["nwalkers"],
            ndim=model.prior_count,
            log_prob_fn=fitness_function.__call__,
            backend=emcee.backends.HDFBackend(filename=self.backend_filename),
            pool=pool,
        )

        try:

            emcee_state = emcee_sampler.get_last_sample()
            samples = self.samples_from(model=model)

            total_iterations = emcee_sampler.iteration

            if samples.converged:
                iterations_remaining = 0
            else:
                iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

                self.logger.info(
                    "Existing Emcee samples found, resuming non-linear search."
                )

        except AttributeError:

            (
                unit_parameter_lists,
                parameter_lists,
                log_posterior_list,
            ) = self.initializer.samples_from_model(
                total_points=emcee_sampler.nwalkers,
                model=model,
                fitness_function=fitness_function,
            )

            emcee_state = np.zeros(shape=(emcee_sampler.nwalkers, model.prior_count))

            self.logger.info("No Emcee samples found, beginning new non-linear search.")

            for index, parameters in enumerate(parameter_lists):
                emcee_state[index, :] = np.asarray(parameters)

            total_iterations = 0
            iterations_remaining = self.config_dict_run["nsteps"]

        while iterations_remaining > 0:

            if self.iterations_per_update > iterations_remaining:
                iterations = iterations_remaining
            else:
                iterations = self.iterations_per_update

            for sample in emcee_sampler.sample(
                initial_state=emcee_state,
                iterations=iterations,
                progress=True,
                skip_initial_state_check=True,
                store=True,
            ):
                pass

            emcee_state = emcee_sampler.get_last_sample()

            total_iterations += iterations
            iterations_remaining = self.config_dict_run["nsteps"] - total_iterations

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=True
            )

            if self.auto_correlations_settings.check_for_convergence:
                if emcee_sampler.iteration > self.auto_correlations_settings.check_size:
                    if samples.converged:
                        iterations_remaining = 0

        self.logger.info("Emcee sampling complete.")

    def config_dict_with_test_mode_settings_from(self, config_dict):

        return {
            **config_dict,
            "nwalkers": 20,
            "nsteps": 10,
        }

    def fitness_function_from_model_and_analysis(
        self, model, analysis, log_likelihood_cap=None
    ):

        return Emcee.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap,
        )

    @property
    def backend_filename(self):
        return path.join(self.paths.samples_path, "emcee.hdf")

    @property
    def backend(self) -> emcee.backends.HDFBackend:
        """
        The `Emcee` hdf5 backend, which provides access to all samples, likelihoods, etc. of the non-linear search.

        The sampler is described in the "Results" section at https://dynesty.readthedocs.io/en/latest/quickstart.html
        """

        if os.path.isfile(self.backend_filename):
            return emcee.backends.HDFBackend(filename=self.backend_filename)
        else:
            raise FileNotFoundError(
                "The file emcee.hdf does not exist at the path "
                + self.paths.samples_path
            )

    def samples_from(self, model):

        return SamplesEmcee.from_results_internal(
            model=model,
            results_internal=self.backend,
            auto_correlation_settings=self.auto_correlations_settings,
            time=self.timer.time,
        )

    def plot_results(self, samples):
        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["emcee"][name]

        plotter = EmceePlotter(
            samples=samples,
            output=Output(
                path=path.join(self.paths.image_path, "search"), format="png"
            ),
        )

        if should_plot("corner"):
            plotter.corner()

        if should_plot("trajectories"):
            plotter.trajectories()

        if should_plot("likelihood_series"):
            plotter.likelihood_series()

        if should_plot("time_series"):
            plotter.time_series()



import logging
from typing import Dict, Tuple, Optional, List

import numpy as np

from autoconf import cached_property
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.graphical.mean_field import MeanField, FactorApproximation
from autofit.graphical.utils import Status
from autofit.mapper.variable import Variable, Plate, VariableData
from autofit.messages.abstract import AbstractMessage

logger = logging.getLogger(__name__)


class EPMeanField(FactorGraph):
    """
    this class encode the EP mean-field approximation to a factor graph


    Attributes
    ----------
    factor_graph: FactorGraph
        the base factor graph being approximated

    factor_mean_field: Dict[Factor, MeanField]
        the mean-field approximation for each factor in the factor graph

    mean_field: MeanField
        the mean-field approximation of the full factor graph
        i.e. the product of the factor mean-field approximations

    variables: Set[Variable]
        the variables of the approximation

    deterministic_variables: Set[Variable]
        the deterministic variables

    log_evidence: float
        the approximate log evidence of the approximation

    is_valid: bool
        returns whether the factor mean-field approximations are all valid

    Methods
    -------
    from_approx_dists(factor_graph, approx_dists)
        create a EPMeanField object from the passed factor_graph
        using approx_dists to initialise the factor mean-field approximations

    factor_approximation(factor)
        create the FactorApproximation for the factor

    project_factor_approx(factor_approximation)
        given the passed FactorApproximation, return a new `EPMeanField`
        object encoding the updated mean-field approximation
    """

    def __init__(
            self, factor_graph: FactorGraph, factor_mean_field: Dict[Factor, MeanField]
    ):
        self._factor_graph = factor_graph
        self._factor_mean_field = factor_mean_field

        super().__init__(self.factor_graph.factors)

    @property 
    def parameters(self):
        return {
            'factor_graph': self._factor_graph, 
            'factor_mean_field': self._factor_mean_field
        }

    def copy(self):
        return type(self)(**{k: val.copy() for k, val in self.parameters.items()})

    def subset(self, plates_index: Dict[Plate, List[int]]) -> "EPMeanFieldSubset":
        """Given a dictionary of Plates with a subset of indexes 
        returns the EPMeanFieldSubset that corresponds to the subset of indexes.
        """
        factor_subset_factor = {}
        factor_mean_field_subset = {}
        factor_mean_field_rescale = {}
        for factor, mean_field in self.factor_mean_field.items():
            plate_sizes = VariableData.plate_sizes(mean_field)
            factor_subset_factor[factor] = subset_factor = factor.subset(
                plates_index, plate_sizes
            )

            mean_field_subset = mean_field.subset(plates_index=plates_index)
            factor_mean_field_subset[subset_factor] = mean_field_subset
            mean_field_size = VariableData.prod(plate_sizes)
            subset_size = VariableData.prod(VariableData.plate_sizes(mean_field_subset))
            scale_factor = subset_size / mean_field_size
            factor_mean_field_rescale[subset_factor] = VariableData({
                v: scale_factor * mean_field[v].size / message.size
                for v, message in mean_field_subset.items()
            })

        subset_factor_graph = FactorGraph(factor_subset_factor.values())
        return EPMeanFieldSubset(
            subset_factor_graph,
            factor_mean_field_subset,
            factor_mean_field_rescale,
            factor_subset_factor,
            self,
            plates_index,
        )

    __getitem__ = subset

    def __setitem__(self, index, subset_approx):
        subset_fac_meanfield = subset_approx.factor_mean_field 
        for factor, subset_fac  in subset_approx._factor_subset_factor.items():
            new_dist = subset_fac_meanfield[subset_fac]
            self.update_factor_mean_field(factor, new_dist, index)

    def update(self, subset_approx):
        self[subset_approx.plates_index] = subset_approx
        return self 

    def merge(self, index, subset_approx):
        factor_mean_field = self.factor_mean_field
        subset_fac_mean_field = subset_approx.factor_mean_field 
        for factor, subset_fac  in subset_approx._factor_subset_factor.items():
            new_dist = subset_fac_mean_field[subset_fac]
            old_dist = factor_mean_field[factor] 
            factor_mean_field[factor] = old_dist.merge(index, new_dist)

        return type(self)(
            self.factor_graph, 
            factor_mean_field, 
        )

    def update_factor_mean_field(
            self,
            factor: Factor,
            new_dist: MeanField,
            index: Optional[Dict[Plate, List[int]]] = None, 
    ):
        if index:
            self._factor_mean_field[factor][index] = new_dist
        else:
            self._factor_mean_field[factor] = new_dist

    @property
    def name(self):
        return f"EP_{self.factor_graph.name}"

    @cached_property
    def fixed_values(self):
        return {
            k: v for mf in self._factor_mean_field.values() 
            for k, v in mf.fixed_values.items()
        }

    @property
    def variables(self):
        return self.factor_graph.variables

    @property
    def deterministic_variables(self):
        return self.factor_graph.deterministic_variables

    @property
    def factor_mean_field(self) -> Dict[Factor, MeanField]:
        return self._factor_mean_field.copy()

    @property
    def factor_graph(self) -> FactorGraph:
        return self._factor_graph

    @classmethod
    def from_approx_dists(
            cls,
            factor_graph: FactorGraph,
            approx_dists: Dict[Variable, AbstractMessage],
    ) -> "EPMeanField":
        factor_mean_field = {
            factor: MeanField({v: approx_dists[v].copy() for v in factor.all_variables})
            for factor in factor_graph.factors
        }

        return cls(factor_graph, factor_mean_field)

    from_kws = from_approx_dists

    def factor_approximation(self, factor: Factor) -> FactorApproximation:
        """
        Create an approximation for one factor.

        This comprises:
        - The factor
        - The factor's variable distributions
        - The cavity distribution, which is the product of the distributions
          for each variable for all other factors
        - The model distribution, which is the product of the distributions
          for each variable for all factors

        Parameters
        ----------
        factor
            Some factor

        Returns
        -------
        An object comprising distributions with a specific distribution excluding
        that factor
        """
        factor_mean_field = self._factor_mean_field.copy()
        factor_dist = factor_mean_field.pop(factor)
        cavity_dist = MeanField({
            v: 1.0
            for v in factor_dist.all_variables
        }).prod(
            *factor_mean_field.values(),
            default=factor_dist,
        )

        model_dist = factor_dist.prod(cavity_dist, default=factor)

        return FactorApproximation(factor, cavity_dist, factor_dist, model_dist)

    get_factor_approx = factor_approximation

    def project_mean_field(
        self,
        new_dist: MeanField,
        factor_approx: FactorApproximation,
        delta: float = 1.0,
        status: Status = Status(),
    ) -> Tuple["EPMeanField", Status]:

        new_factor_dist, status = new_dist.update_factor_mean_field(
            factor_approx.cavity_dist,
            factor_approx.factor_dist,
            delta=delta,
            status=status,
        )
        factor_mean_field = self.factor_mean_field
        factor_mean_field[factor_approx.factor] = new_factor_dist

        new_approx = type(self)(
            factor_graph=self._factor_graph, 
            factor_mean_field=factor_mean_field,
        )
        return new_approx, status

    def project_factor_approx(
            self,
            projection: FactorApproximation,
            status: Optional[Status] = None,
    ) -> Tuple["EPMeanField", Status]:
        """ """
        factor_mean_field = self.factor_mean_field
        factor_mean_field[projection.factor] = projection.factor_dist

        new_approx = type(self)(
            factor_graph=self._factor_graph, 
            factor_mean_field=factor_mean_field
        )
        return new_approx, status

    project = project_factor_approx

    @property
    def mean_field(self) -> MeanField:
        return MeanField({v: 1.0 for v in self.all_variables}).prod(
            *self._factor_mean_field.values()
        )

    model_dist = mean_field

    @property
    def variable_messages(self) -> Dict[Variable, List[AbstractMessage]]:
        variable_messages = {v: [] for v in self.all_variables}
        for meanfield in self.factor_mean_field.values():
            for v, message in meanfield.items():
                variable_messages[v].append(message)

        return variable_messages

    @property
    def variable_message_count(self) -> Dict[Variable, int]:
        """
        The number of messages (i.e. the number of factors) associated with each
        variable
        """
        return {
            variable: len(messages)
            for variable, messages
            in self.variable_messages.items()
        }

    @property
    def variable_evidence(self) -> Dict[Variable, np.ndarray]:
        return {
            v: AbstractMessage.log_normalisation(*ms)
            for v, ms in self.variable_messages.items()
        }

    @property
    def factor_evidence(self) -> Dict[Factor, np.ndarray]:
        return {
            factor: meanfield.log_norm
            for factor, meanfield in self.factor_mean_field.items()
        }

    @property
    def log_evidence(self):
        """
        Calculates evidence for the EP approximation

        Evidence for a variable, xᵢ,

        Zᵢ = ∫ ∏ₐ m_{a → i} (xᵢ) dxᵢ

        Evidence for a factor, f_a,

                ∫ ∏_{j ∈ a} m_{i → a} (xᵢ) fₐ(xₐ) dxₐ
        Zₐ = -----------------------------------------
                             ∏_{j ∈ a} Zⱼ

        Evidence for model

        Z = ∏ᵢ Zᵢ ∏ₐ Zₐ
        """
        variable_evidence = {
            v: np.sum(logz) for v, logz in self.variable_evidence.items()
        }
        factor_evidence = sum(
            np.sum(meanfield.log_norm)
            - sum(variable_evidence[v] for v in factor.all_variables)
            for factor, meanfield in self.factor_mean_field.items()
        )
        return factor_evidence + sum(variable_evidence.values())

    def __repr__(self) -> str:
        clsname = type(self).__name__
        try:
            log_evidence = self.log_evidence
        except Exception as e:
            logger.exception(e)
            log_evidence = float("nan")
        return f"{clsname}({self.factor_graph}, " f"log_evidence={log_evidence})"


class EPMeanFieldSubset(EPMeanField):
    """
    this class encode the EP mean-field approximation to a factor graph


    Attributes
    ----------
    factor_graph: FactorGraph
        the base factor graph being approximated

    factor_mean_field: Dict[Factor, MeanField]
        the mean-field approximation for each factor in the factor graph

    mean_field: MeanField
        the mean-field approximation of the full factor graph
        i.e. the product of the factor mean-field approximations

    variables: Set[Variable]
        the variables of the approximation

    deterministic_variables: Set[Variable]
        the deterministic variables

    log_evidence: float
        the approximate log evidence of the approximation

    is_valid: bool
        returns whether the factor mean-field approximations are all valid

    Methods
    -------
    from_approx_dists(factor_graph, approx_dists)
        create a EPMeanField object from the passed factor_graph
        using approx_dists to initialise the factor mean-field approximations

    factor_approximation(factor)
        create the FactorApproximation for the factor

    project_factor_approx(factor_approximation)
        given the passed FactorApproximation, return a new `EPMeanField`
        object encoding the updated mean-field approximation
    """

    def __init__(
        self,
        factor_graph: FactorGraph,
        factor_mean_field: Dict[Factor, MeanField],
        factor_rescale: Dict[Factor, Dict[Variable, float]],
        factor_subset_factor: Dict[Factor, Factor],
        ep_mean_field: EPMeanField,
        plates_index: Dict[Plate, List[int]],
    ):
        super().__init__(factor_graph, factor_mean_field)

        self._factor_rescale = factor_rescale
        self._factor_subset_factor = factor_subset_factor
        self._ep_mean_field = ep_mean_field
        self._plates_index = plates_index

    
    @property 
    def parameters(self):
        return {
            "factor_graph": self._factor_graph, 
            "factor_mean_field": self._factor_mean_field,
            "factor_rescale": self._factor_rescale,
            "factor_subset_factor": self._factor_subset_factor,
            "ep_mean_field": self._ep_mean_field,
            "plates_index": self._plates_index,
        }

    @property 
    def plates_index(self):
        return self._plates_index.copy()

    @property 
    def factor_rescale(self):
        return self._factor_rescale.copy()

    @property 
    def factor_subset_factor(self):
        return self._factor_subset_factor.copy()

    def project_mean_field(
        self,
        new_model_dist: MeanField,
        factor_approx: FactorApproximation,
        delta: float = 1.0,
        status: Status = Status(),
    ) -> Tuple["EPMeanField", Status]:

        factor_mean_field = self.factor_mean_field
        
        # We're fitting the full factor_dist, not the subset factor_dist
        rescale1 = {
            v: 1 - scale 
            for v, scale in self._factor_rescale[factor_approx.factor].items()
        }
        last_factor_dist = factor_mean_field.pop(factor_approx.factor)
        subset_cavity_dist = last_factor_dist.rescale(rescale1)
        
        if delta < 1:
            # This is a more stable update that effectively performs an
            # Exponential moving average of the natural parameters of the
            # factor message with delta as the EMA coefficient
            factor_mean_field[factor_approx.factor], status = (
                new_model_dist * subset_cavity_dist
            ).update_factor_mean_field(
                factor_approx.cavity_dist,
                factor_approx.factor_dist * subset_cavity_dist,
                delta=delta,
                status=status,
            )
        else:
            new_factor_dist, status = new_model_dist.update_factor_mean_field(
                factor_approx.cavity_dist,
                factor_approx.factor_dist,
                delta=delta,
                status=status,
            )
            factor_mean_field[factor_approx.factor] = new_factor_dist  * subset_cavity_dist
            
        new_approx = type(self)(
            factor_graph=self._factor_graph,
            factor_mean_field=factor_mean_field,
            factor_rescale=self._factor_rescale,
            factor_subset_factor=self._factor_subset_factor, 
            ep_mean_field=self._ep_mean_field,
            plates_index=self._plates_index,
        )
        return new_approx, status

    def factor_approximation(self, factor: Factor) -> FactorApproximation:
        """
        Create an approximation for one factor.

        This comprises:
        - The factor
        - The factor's variable distributions
        - The cavity distribution, which is the product of the distributions
          for each variable for all other factors
        - The model distribution, which is the product of the distributions
          for each variable for all factors

        Parameters
        ----------
        factor
            Some factor

        Returns
        -------
        An object comprising distributions with a specific distribution excluding
        that factor
        """
        factor = self._factor_subset_factor.get(factor, factor)

        factor_mean_field = self._factor_mean_field.copy()

        factor_dist = factor_mean_field.pop(factor)
        cavity_dist = dict(
            MeanField({v: 1.0 for v in factor_dist.all_variables}).prod(
                *factor_mean_field.values()
            )
        )
        model_dist = factor_dist.prod(cavity_dist)

        factor_dist = dict(factor_dist)
        for v, scale in self._factor_rescale[factor].items():
            if scale < 1:
                cavity_dist[v] = cavity_dist[v] * factor_dist[v] ** (1 - scale)
                factor_dist[v] = factor_dist[v] ** scale

        return FactorApproximation(
            factor, MeanField(cavity_dist), MeanField(factor_dist), model_dist
        )



import csv
from copy import copy
from typing import List, Tuple

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.tools.util import split_paths


class Sample:
    def __init__(
            self,
            log_likelihood: float,
            log_prior: float,
            weight: float,
            kwargs=None
    ):
        """
        One sample taken during a search

        Parameters
        ----------
        log_likelihood
            The likelihood associated with this instance
        log_prior
            A logarithmic prior of the instance
        weight
            The weight this sample contributes to the PDF.
        kwargs
            Dictionary mapping model paths to values for the sample
        """
        self.log_likelihood = log_likelihood
        self.log_prior = log_prior
        self.weight = weight
        self.kwargs = kwargs or dict()

    @property
    def log_posterior(self) -> float:
        """
        Compute the posterior
        """
        return self.log_likelihood + self.log_prior

    def parameter_lists_for_model(
            self,
            model: AbstractPriorModel
    ) -> List[float]:
        """
        Values for instantiating a model, in the same order as priors
        from the model.

        Parameters
        ----------
        model
            The model from which this was a sample

        Returns
        -------
        A list of physical values
        """
        if self.is_path_kwargs:
            paths = model.all_paths
        else:
            paths = model.all_names

        return self.parameter_lists_for_paths(
            paths
        )

    def parameter_lists_for_paths(
            self,
            paths
    ):
        result = list()
        for keys in paths:
            is_found = False
            for key in keys:
                if key in self.kwargs:
                    result.append(
                        self.kwargs[
                            key
                        ]
                    )
                    is_found = True
                    break
            if not is_found:
                raise KeyError(
                    f"Could not find any of the following keys in kwargs {keys}"
                )
        return result

    @property
    def is_path_kwargs(self) -> bool:
        """
        Are the keys in the kwargs dictionary tuples? If they
        are this indicates that they are explicit paths through
        the model.
        """
        for key in self.kwargs.keys():
            return isinstance(
                key, tuple
            )
        return False

    def subsample(self, path_map):
        arg_dict = {}
        for paths, new_path in path_map.items():
            for path in paths:
                if path in self.kwargs:
                    arg_dict[new_path] = self.kwargs[path]
                    break
            else:
                raise KeyError(f"No path from {paths} in sample")

        return type(self)(
            log_likelihood=self.log_likelihood,
            log_prior=self.log_prior,
            weight=self.weight,
            kwargs=arg_dict
        )

    @classmethod
    def from_lists(
            cls,
            model: AbstractPriorModel,
            parameter_lists: List[List[float]],
            log_likelihood_list: List[float],
            log_prior_list: List[float],
            weight_list: List[float]
    ) -> List["Sample"]:
        """
        Convenience method to create a list of samples from lists of contained values
        """
        samples = list()

        paths = model.unique_prior_paths

        for params, log_likelihood, log_prior, weight in zip(
                parameter_lists,
                log_likelihood_list,
                log_prior_list,
                weight_list
        ):
            arg_dict = {
                t: param
                for t, param
                in zip(
                    paths,
                    params
                )
            }

            samples.append(
                cls(
                    log_likelihood=log_likelihood,
                    log_prior=log_prior,
                    weight=weight,
                    kwargs=arg_dict
                )
            )
        return samples

    def instance_for_model(self, model: AbstractPriorModel):
        """
        Create an instance from this sample for a model

        Parameters
        ----------
        model
            The model the this sample was taken from

        Returns
        -------
        The instance corresponding to this sample
        """
        try:
            if self.is_path_kwargs:
                return model.instance_from_path_arguments(
                    self.kwargs
                )
            else:
                return model.instance_from_prior_name_arguments(
                    self.kwargs
                )

        except KeyError:
            # TODO: Does this get used? If so, why?
            return model.instance_from_vector(
                self.parameter_lists_for_model(model)
            )

    @split_paths
    def with_paths(
            self,
            paths: List[Tuple[str, ...]]
    ) -> "Sample":
        """
        Create a copy of this object retaining only the kwargs for which
        there is a matching path in paths.

        Parameters
        ----------
        paths
            A list of paths for which attributes should be retained.

        Returns
        -------
        A reduced sample
        """
        with_paths = copy(self)
        with_paths.kwargs = {
            key: value
            for key, value
            in self.kwargs.items()
            if any(
                all(
                    first == second
                    for first, second
                    in zip(
                        key, path
                    )
                )
                for path in paths
            )
        }
        return with_paths

    @split_paths
    def without_paths(
            self,
            paths: List[Tuple[str, ...]]
    ) -> "Sample":
        """
        Create a copy of this object retaining only the kwargs for which
        there is no matching path in paths.

        Parameters
        ----------
        paths
            A list of paths for which attributes should be removed.

        Returns
        -------
        A reduced sample
        """
        without_paths = copy(self)
        without_paths.kwargs = {
            key: value
            for key, value
            in self.kwargs.items()
            if not any(
                all(
                    first == second
                    for first, second
                    in zip(
                        key, path
                    )
                )
                for path in paths
            )
        }
        return without_paths


def load_from_table(filename: str) -> List[Sample]:
    """
    Load samples from a table

    Parameters
    ----------
    filename
        The path to a CSV file

    Returns
    -------
    A list of samples, one for each row in the CSV
    """
    samples = list()

    sample_args = (
        "log_likelihood",
        "log_prior",
        "weight",
    )

    with open(filename, "r+", newline="") as f:
        reader = csv.reader(f)
        headers = next(reader)
        for row in reader:
            d = {
                header: float(value)
                for header, value
                in zip(
                    headers,
                    row
                )
            }

            samples.append(
                Sample(
                    **{
                        key: value
                        for key, value
                        in d.items()
                        if key in sample_args
                    },
                    kwargs={
                        key: value
                        for key, value
                        in d.items()
                        if key not in sample_args
                    }
                )
            )

    return samples



import inspect
from numbers import Real
from typing import Optional, Set, cast

from autofit.database.query import condition as c
from autofit.database.query.junction import AbstractJunction
from .abstract import AbstractQuery


def _make_comparison(
        symbol: str,
        other
) -> c.AbstractCondition:
    """
    Create the appropriate comparison class, depending on the type of other.

    Parameters
    ----------
    symbol
        The symbol by which the condition will compare
    other
        Some object to be compared to entries in the database

    Returns
    -------
    A condition
    """
    if other is None:
        if symbol != "=":
            raise AssertionError(
                "Inequalities to None do not make sense"
            )
        return c.NoneCondition()

    if isinstance(other, str):
        return c.StringValueCondition(
            symbol, other
        )
    if isinstance(other, Real):
        return c.ValueCondition(
            symbol, other
        )
    if inspect.isclass(other):
        if symbol != "=":
            raise AssertionError(
                "Inequalities to types do not make sense"
            )
        return c.TypeCondition(
            other
        )
    raise AssertionError(
        f"Cannot evaluate equality to type {type(other)}"
    )


class NamedQuery(AbstractQuery):
    def __init__(
            self,
            name: str,
            condition: Optional[c.AbstractCondition] = None,
            inverted=False
    ):
        """
        An object which can be converted into SQL and used to query the database.

        Model instances are flat packed into the database using the name of an
        attribute and a foreign key of the parent's id to express the original
        parent - attribute relationships.

        Queries can be constructed on these models.

        e.g.
        aggregator.galaxies.lens == al.lp.SersicLightProfile

        This class is responsible for the construction of queries from this syntax.
        The name is the name in the path, e.g. galaxies.

        Parameters
        ----------
        name
            The name of an attribute that is stored in the database
        condition
            A child condition combined with the name to produce a query
        """
        super().__init__(condition)
        self.name = name
        self._inverted = inverted

    def __invert__(self):
        return NamedQuery(
            name=self.name,
            condition=self._condition,
            inverted=not self._inverted
        )

    @property
    def other_condition(self) -> c.AbstractCondition:
        return self._condition

    @property
    def condition(self) -> c.AbstractCondition:
        """
        The combined condition that forms the WHERE statement of the query.

        This is at least a check on the name in the object table, but may
        include subqueries of arbitrary complexity expressed by 'other_condition'
        """
        condition = c.NameCondition(
            self.name
        )
        if super().condition:
            condition &= super().condition
        return condition

    @property
    def tables(self) -> Set[c.Table]:
        """
        The set of tables used in the FROM component of the query
        """
        return self.condition.tables

    def __repr__(self):
        return self.query

    def __call__(self, *args, **kwargs):
        raise AttributeError(
            f"'Aggregator' object has no attribute '{self.name}'"
        )

    @property
    def query(self) -> str:
        """
        The SQL string produced by this query. This is applied directly to the database.
        """
        return f"SELECT parent_id FROM {self.tables_string} WHERE {self.condition}"

    @property
    def _in(self):
        if self._inverted:
            return f"NOT IN"
        return "IN"

    @property
    def fit_query(self) -> str:
        return f"SELECT id FROM fit WHERE instance_id {self._in} ({self.query})"

    @property
    def tables_string(self) -> str:
        """
        A string found in the FROM component. Includes aliasing and join
        statements if there is more than one table.

        Currently it is assumed that only the object table and up to one
        further table are required to execute any given query.
        """
        tables = sorted(self.tables)
        first = tables[0]

        string = str(first)

        if len(tables) == 2:
            second = tables[1]
            string = f"{string} JOIN {second} ON {first.abbreviation}.id = {second.abbreviation}.id"
        if len(tables) > 2:
            raise AssertionError(
                "Currently maximum of 2 tables supported"
            )

        return string

    def __str__(self):
        return f"o.id {self._in} ({self.query})"

    def __getattr__(self, item: str):
        """
        Used to extend the query.

        Supports the syntax:
        galaxies.lens.normalization

        Child queries are recursively searched until one with a None child is found.
        A new query is constructed to replace this None, extending the linked list
        of Named queries.

        If any child query is not a Named query then an exception is thrown as this
        does not make sense.

        For example, (one.two.three == 1).four does not make sense.

        Parameters
        ----------
        item
            The name of an attribute

        Returns
        -------
        A newly created query that is the same as this query but with an additional
        NamedQuery added on the end for the new attribute.
        """
        if item == "__setstate__":
            raise AttributeError()

        if self.other_condition is None:
            return NamedQuery(
                self.name,
                NamedQuery(
                    item
                )
            )

        if isinstance(
                self.other_condition,
                NamedQuery
        ):
            return NamedQuery(
                self.name,
                getattr(
                    self.other_condition,
                    item
                )
            )

        raise AttributeError(
            "Can only extend a simple path query"
        )

    def __ne__(self, other):
        return ~(self == other)

    def __eq__(self, other):
        if isinstance(other, NamedQuery):
            return other.query == self.query

        return self._recursive_comparison(
            "=",
            other
        )

    def __gt__(self, other):
        if isinstance(other, c.AbstractCondition):
            return super().__gt__(other)
        return self._recursive_comparison(
            ">",
            other
        )

    def __ge__(self, other):
        return self._recursive_comparison(
            ">=",
            other
        )

    def __lt__(self, other):
        if isinstance(other, c.AbstractCondition):
            return super().__lt__(other)
        return self._recursive_comparison(
            "<",
            other
        )

    def __le__(self, other):
        return self._recursive_comparison(
            "<=",
            other
        )

    def __hash__(self):
        return hash(str(self))

    def _recursive_comparison(
            self,
            symbol: str,
            other
    ) -> "NamedQuery":
        """
        Create a new NamedQuery, recursing through each sub NamedQuery until
        one without a child query is found. The final NamedQuery is given a
        child query.

        This supports the syntax:
        aggregator.galaxies.centre.normalization >= 1

        The final query added on the end is the comparison query ">= 1"

        Parameters
        ----------
        symbol
            =, >=, <=, >, <
        other
            An object to compare to values in the database.

        Returns
        -------
        A new NamedQuery with the comparison query added on the end
        """
        if self.other_condition is None:
            return NamedQuery(
                self.name,
                _make_comparison(
                    symbol,
                    other
                )
            )

        if isinstance(
                self.other_condition,
                NamedQuery
        ):
            return NamedQuery(
                self.name,
                cast(
                    NamedQuery,
                    self.other_condition
                )._recursive_comparison(
                    symbol,
                    other
                )
            )

        if isinstance(
                self._condition,
                AbstractJunction
        ):
            raise AssertionError(
                "Cannot compare a complex query"
            )

        raise AssertionError(
            f"Cannot evaluate equality to type {type(other)}"
        )



import logging
from abc import ABC, abstractmethod
from typing import Tuple

from autofit.graphical.mean_field import MeanField, FactorApproximation, Status
from autofit.graphical.utils import LogWarnings

logger = logging.getLogger(__name__)


class AbstractFactorOptimiser(ABC):
    """
    An optimiser used to optimise individual factors during EPOptimisation.
    """

    logger = logger.debug

    def __init__(
        self, initial_values=None, inplace=False,
    ):
        self.initial_values = initial_values or {}
        self.inplace = inplace

    @abstractmethod
    def optimise(
        self, factor_approx: FactorApproximation, status: Status = Status()
    ) -> Tuple[MeanField, Status]:
        pass

    def exact_fit(
        self, factor_approx: FactorApproximation, status: Status = Status()
    ) -> Tuple[MeanField, Status]:
        factor = factor_approx.factor
        cavity_dist = factor_approx.cavity_dist
        with LogWarnings(logger=self.logger, action="always") as caught_warnings:
            if factor._calc_exact_update:
                factor_mean_field = factor.calc_exact_update(cavity_dist)
                new_model_dist = factor_mean_field * cavity_dist
            elif factor._calc_exact_projection:
                new_model_dist = factor.calc_exact_projection(cavity_dist)
            else:
                raise NotImplementedError("Factor does not have exact updates methods")

        return new_model_dist, status


class ExactFactorFit(AbstractFactorOptimiser):
    optimise = AbstractFactorOptimiser.exact_fit



from dynesty.results import Results
from dynesty import utils as dyfunc
import numpy as np
from typing import Optional

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples import Sample
from autofit.non_linear.samples.nest import SamplesNest


class SamplesDynesty(SamplesNest):

    def __add__(
            self,
            other: "SamplesDynesty"
    ) -> "SamplesDynesty":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        For dynesty samples, the in-built dynesty function `merge_runs` can be used to combine results_internal in their native
        format and therefore retain visualization support.

        Parameters
        ----------
        other
            Another Samples class

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        results_internal = dyfunc.merge_runs(res_list=[self.results_internal, other.results_internal])

        return SamplesDynesty(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
            number_live_points=self._number_live_points,
            unconverged_sample_size=self.unconverged_sample_size,
            time=self.time,
            results_internal=results_internal
        )

    @classmethod
    def from_results_internal(
            cls,
            results_internal: Results,
            model: AbstractPriorModel,
            number_live_points: int,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results of a `NonLinearSearch` (e.g.
        as files on your hard-disk) and Python.

        To create a `Samples` object after a `dynesty` model-fit the results must be converted from the
        native format used by `dynesty` to lists of values, the format used by the **PyAutoFit** `Samples` objects.

        This classmethod performs this conversion before creating a `SamplesDynesty` object.

        Parameters
        ----------
        results_internal
            The `dynesty` results in their native internal format from which the samples are computed.
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        number_live_points
            The number of live points used by the `dynesty` search.
        unconverged_sample_size
            If the samples are for a search that is yet to convergence, a reduced set of samples are used to provide
            a rough estimate of the parameters. The number of samples is set by this parameter.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        """
        parameter_lists = results_internal.samples.tolist()
        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)
        log_likelihood_list = list(results_internal.logl)

        try:
            weight_list = list(
                np.exp(np.asarray(results_internal.logwt) - results_internal.logz[-1])
            )
        except:
            weight_list = results_internal["weights"]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list,
        )

        return SamplesDynesty(
            model=model,
            sample_list=sample_list,
            number_live_points=number_live_points,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
            results_internal=results_internal,
        )

    @property
    def number_live_points(self):
        return self._number_live_points

    @property
    def total_samples(self):
        return int(np.sum(self.results_internal.ncall))

    @property
    def log_evidence(self):
        return np.max(self.results_internal.logz)



import math

import numpy as np
from scipy import special

from autoconf import cached_property
from autofit.messages.abstract import AbstractMessage
from autofit.messages.utils import invpsilog


class GammaMessage(AbstractMessage):
    @property
    def log_partition(self):
        alpha, beta = GammaMessage.invert_natural_parameters(self.natural_parameters)
        return special.gammaln(alpha) - alpha * np.log(beta)

    log_base_measure = 0.0
    _support = ((0, np.inf),)
    _parameter_support = ((0, np.inf), (0, np.inf))

    def __init__(
            self,
            alpha=1.0,
            beta=1.0,
            lower_limit=-math.inf,
            upper_limit=math.inf,
            log_norm=0.0,
            id_=None
    ):
        self.alpha = alpha
        self.beta = beta
        super().__init__(
            alpha,
            beta,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            log_norm=log_norm,
            id_=id_
        )

    def value_for(self, unit: float) -> float:
        raise NotImplemented()

    @cached_property
    def natural_parameters(self):
        return self.calc_natural_parameters(self.alpha, self.beta)

    @staticmethod
    def calc_natural_parameters(alpha, beta):
        return np.array([alpha - 1, -beta])

    @staticmethod
    def invert_natural_parameters(natural_parameters):
        eta1, eta2 = natural_parameters
        return eta1 + 1, -eta2

    @staticmethod
    def to_canonical_form(x):
        return np.array([np.log(x), x])

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats):
        logX, X = suff_stats
        alpha = invpsilog(logX - np.log(X))
        beta = alpha / X
        return cls.calc_natural_parameters(alpha, beta)

    @cached_property
    def mean(self):
        return self.alpha / self.beta

    @cached_property
    def variance(self):
        return self.alpha / self.beta ** 2

    def sample(self, n_samples=None):
        a1, b1 = self.parameters
        shape = (n_samples,) + self.shape if n_samples else self.shape
        return np.random.gamma(a1, scale=1 / b1, size=shape)

    @classmethod
    def from_mode(cls, mode, covariance, **kwargs):
        m, V = cls._get_mean_variance(mode, covariance)

        alpha = 1 + m ** 2 * V  # match variance
        beta = alpha / m  # match mean
        return cls(alpha, beta, **kwargs)

    def kl(self, dist):
        P, Q = dist, self
        logP = np.log(P.alpha)
        # TODO check this is correct
        # https://arxiv.org/pdf/0911.4863.pdf
        return (
                (P.alpha - Q.alpha) * special.psi(P.alpha)
                - special.gammaln(P.alpha)
                + special.gammaln(Q.alpha)
                + Q.alpha * (np.log(P.beta / Q.beta))
                + P.alpha * (Q.beta / P.beta - 1)
        )

    def logpdf_gradient(self, x):
        logl = self.logpdf(x)
        eta1 = self.natural_parameters[0]
        gradl = eta1 / x - self.beta
        return logl, gradl

    def logpdf_gradient_hessian(self, x):
        logl = self.logpdf(x)
        eta1 = self.natural_parameters[0]
        gradl = eta1 / x
        hessl = -gradl / x
        gradl -= self.beta
        return logl, gradl, hessl



import abc
import pickle

from dill import register

from .non_linear.grid.grid_search import GridSearch as SearchGridSearch
from . import conf
from . import exc
from . import mock as m
from .database.aggregator.aggregator import GridSearchAggregator
from .graphical.expectation_propagation.history import EPHistory
from .graphical.declarative.factor.analysis import AnalysisFactor
from .graphical.declarative.collection import FactorGraphModel
from .graphical.declarative.factor.hierarchical import HierarchicalFactor
from .graphical.laplace import LaplaceOptimiser
from .non_linear.samples import SamplesMCMC
from .non_linear.samples import SamplesNest
from .non_linear.samples import Samples
from .non_linear.samples import SamplesPDF
from .non_linear.samples import Sample
from .non_linear.samples import load_from_table
from .non_linear.samples import SamplesStored
from .database.aggregator import Aggregator
from .database.model import Fit
from .database.aggregator import Query
from .database.model.fit import Fit
from .aggregator.search_output import SearchOutput
from .mapper import prior
from .mapper.model import AbstractModel
from .mapper.model import ModelInstance
from .mapper.model import ModelInstance as Instance
from .mapper.model import path_instances_of_class
from .mapper.model_mapper import ModelMapper
from .mapper.model_mapper import ModelMapper as Mapper
from .mapper.model_object import ModelObject
from .mapper.operator import DiagonalMatrix
from .mapper.prior.arithmetic.assertion import ComparisonAssertion
from .mapper.prior.arithmetic.assertion import ComparisonAssertion
from .mapper.prior.arithmetic.assertion import GreaterThanLessThanAssertion
from .mapper.prior.arithmetic.assertion import GreaterThanLessThanEqualAssertion
from .mapper.prior.deferred import DeferredArgument
from .mapper.prior.deferred import DeferredInstance
from .mapper.prior.width_modifier import AbsoluteWidthModifier
from .mapper.prior.width_modifier import RelativeWidthModifier
from .mapper.prior.width_modifier import WidthModifier
from .mapper.prior import GaussianPrior
from .mapper.prior import LogGaussianPrior
from .mapper.prior import LogUniformPrior
from .mapper.prior.abstract import Prior
from .mapper.prior.tuple_prior import TuplePrior
from .mapper.prior import UniformPrior
from .mapper.prior_model.abstract import AbstractPriorModel
from .mapper.prior_model.annotation import AnnotationPriorModel
from .mapper.prior_model.attribute_pair import AttributeNameValue
from .mapper.prior_model.attribute_pair import InstanceNameValue
from .mapper.prior_model.attribute_pair import PriorNameValue
from .mapper.prior_model.attribute_pair import cast_collection
from .mapper.prior_model.collection import Collection
from .mapper.prior_model.prior_model import Model
from .mapper.prior_model.prior_model import Model
from .mapper.prior_model.util import PriorModelNameValue
from .non_linear.abstract_search import NonLinearSearch
from .non_linear.abstract_search import PriorPasser
from .non_linear.analysis.analysis import Analysis
from .non_linear.grid.grid_search import GridSearchResult
from .non_linear.initializer import InitializerBall
from .non_linear.initializer import InitializerPrior
from .non_linear.initializer import SpecificRangeInitializer
from .non_linear.mcmc.auto_correlations import AutoCorrelationsSettings
from .non_linear.mcmc.emcee.emcee import Emcee
from .non_linear.mcmc.zeus.zeus import Zeus
from .non_linear.nest.dynesty import DynestyDynamic
from .non_linear.nest.dynesty import DynestyStatic
from .non_linear.nest.ultranest.ultranest import UltraNest
from .non_linear.optimize.drawer.drawer import Drawer
from .non_linear.optimize.lbfgs.lbfgs import LBFGS
from .non_linear.optimize.pyswarms.globe import PySwarmsGlobal
from .non_linear.optimize.pyswarms.local import PySwarmsLocal
from .non_linear.paths import DirectoryPaths
from .non_linear.paths import DatabasePaths
from .non_linear.result import Result
from .non_linear.result import ResultsCollection
from .non_linear.settings import SettingsSearch
from .non_linear.samples.pdf import marginalize
from .example.model import Gaussian
from .text import formatter
from .text import samples_text
from .interpolator import LinearInterpolator, SplineInterpolator
from .tools import util


from autofit.mapper.prior.arithmetic.compound import SumPrior as Add
from autofit.mapper.prior.arithmetic.compound import MultiplePrior as Multiply
from autofit.mapper.prior.arithmetic.compound import DivisionPrior as Divide
from autofit.mapper.prior.arithmetic.compound import ModPrior as Mod
from autofit.mapper.prior.arithmetic.compound import PowerPrior as Power
from autofit.mapper.prior.arithmetic.compound import AbsolutePrior as Abs
from autofit.mapper.prior.arithmetic.compound import Log
from autofit.mapper.prior.arithmetic.compound import Log10

from . import example as ex
from . import database as db


@register(abc.ABCMeta)
def save_abc(pickler, obj):
    pickle._Pickler.save_type(pickler, obj)


conf.instance.register(__file__)

__version__ = "2023.3.21.4"



from .model import Object
from ..sqlalchemy_ import sa


class NoneInstance(Object):
    __tablename__ = "none"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    __mapper_args__ = {
        'polymorphic_identity': 'none'
    }

    def __call__(self):
        return None


class Collection(Object):
    """
    A tuple or list
    """

    __tablename__ = "collection"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True
    )

    __mapper_args__ = {
        'polymorphic_identity': 'collection'
    }

    @classmethod
    def _from_object(
            cls,
            source
    ):
        instance = cls()
        instance.cls = type(source)
        instance._add_children([
            (str(i), item)
            for i, item in enumerate(
                source
            )
        ])
        return instance

    def __call__(self):
        return self.cls([
            child()
            for child
            in sorted(
                self.children,
                key=lambda child: int(child.name)
            )
        ])


class Instance(Object):
    """
    An instance, such as a class instance
    """

    __tablename__ = "instance"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    __mapper_args__ = {
        'polymorphic_identity': 'instance'
    }

    @classmethod
    def _from_object(
            cls,
            source
    ):
        instance = cls()
        instance.cls = type(source)
        if hasattr(
                source,
                "__getstate__"
        ):
            instance._add_children(
                source.__getstate__().items()
            )
        elif hasattr(
                source,
                "__dict__"
        ):
            instance._add_children(
                source.__dict__.items()
            )
        return instance


class Value(Object):
    """
    A float
    """

    __tablename__ = "value"

    __mapper_args__ = {
        'polymorphic_identity': 'value'
    }

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    value = sa.Column(sa.Float)

    @classmethod
    def _from_object(
            cls,
            source
    ):
        instance = cls()
        instance.value = source
        return instance

    def __call__(self):
        return self.value


class StringValue(Object):
    """
    A string
    """

    __tablename__ = "string_value"

    __mapper_args__ = {
        'polymorphic_identity': 'string_value'
    }

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    value = sa.Column(sa.String)

    @classmethod
    def _from_object(
            cls,
            source
    ):
        instance = cls()
        instance.value = source
        return instance

    def __call__(self):
        return self.value



import math
from abc import ABC, abstractmethod
from copy import copy
from functools import reduce
from inspect import getfullargspec
from itertools import count
from numbers import Real
from operator import and_
from typing import Dict, Tuple, Iterator
from typing import Optional, Union, Type, List

import numpy as np

from autoconf import cached_property
from ..mapper.variable import Variable

from .interface import MessageInterface

enforce_id_match = True


def update_array(arr1, ind, arr2):
    if np.shape(arr1):
        out = arr1.copy()
        out[ind] = arr2
        return out

    return arr2


class AbstractMessage(MessageInterface, ABC):

    _Base_class: Optional[Type["AbstractMessage"]] = None
    _projection_class: Optional[Type["AbstractMessage"]] = None
    _multivariate: bool = False
    _parameter_support: Optional[Tuple[Tuple[float, float], ...]] = None
    _support: Optional[Tuple[Tuple[float, float], ...]] = None

    ids = count()

    def __init__(
        self,
        *parameters: Union[np.ndarray, float],
        log_norm=0.0,
        lower_limit=-math.inf,
        upper_limit=math.inf,
        id_=None,
    ):
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.id = next(self.ids) if id_ is None else id_
        self.log_norm = log_norm
        self._broadcast = np.broadcast(*parameters)

        if self.shape:
            self.parameters = tuple(np.asanyarray(p) for p in parameters)
        else:
            self.parameters = tuple(parameters)

    @property
    def broadcast(self):
        return self._broadcast

    def check_support(self) -> np.ndarray:
        if self._parameter_support is not None:
            return reduce(
                and_,
                (
                    (p >= support[0]) & (p <= support[1])
                    for p, support in zip(self.parameters, self._parameter_support)
                ),
            )
        elif self.ndim:
            return np.array(True, dtype=bool, ndmin=self.ndim)
        return np.array([True])

    @property
    def multivariate(self):
        return self._multivariate

    def copy(self):
        cls = self._Base_class or type(self)
        result = cls(
            *(copy(params) for params in self.parameters),
            log_norm=self.log_norm,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
        )
        result.id = self.id
        return result

    def __bool__(self):
        return True

    @abstractmethod
    def sample(self, n_samples: Optional[int] = None):
        pass

    @staticmethod
    @abstractmethod
    def invert_natural_parameters(
        natural_parameters: np.ndarray,
    ) -> Tuple[np.ndarray, ...]:
        pass

    @cached_property
    @abstractmethod
    def variance(self) -> np.ndarray:
        pass

    @cached_property
    def scale(self) -> np.ndarray:
        return self.std

    @cached_property
    def std(self) -> np.ndarray:
        return self.variance ** 0.5

    def __hash__(self):
        return self.id

    def __iter__(self) -> Iterator[np.ndarray]:
        return iter(self.parameters)

    @classmethod
    def _cached_attrs(cls):
        for n in dir(cls):
            attr = getattr(cls, n)
            if isinstance(attr, cached_property):
                yield n

    def _reset_cache(self):
        for attr in self._cached_attrs():
            self.__dict__.pop(attr, None)

    def __getitem__(self, index) -> "AbstractMessage":
        cls = self._Base_class or type(self)
        if index == ():
            return self
        else:
            return cls(*(param[index] for param in self.parameters))

    def __setitem__(self, index, value):
        self._reset_cache()
        for param0, param1 in zip(self.parameters, value.parameters):
            param0[index] = param1

    def merge(self, index, value):
        cls = self._Base_class or type(self)
        return cls(
            *(
                update_array(param0, index, param1)
                for param0, param1 in zip(self.parameters, value.parameters)
            )
        )

    @classmethod
    def from_natural_parameters(
        cls, natural_parameters: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        cls_ = cls._projection_class or cls._Base_class or cls
        args = cls_.invert_natural_parameters(natural_parameters)
        return cls_(*args, **kwargs)

    @classmethod
    @abstractmethod
    def invert_sufficient_statistics(
        cls, sufficient_statistics: np.ndarray
    ) -> np.ndarray:
        pass

    @classmethod
    def from_sufficient_statistics(
        cls, suff_stats: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        natural_params = cls.invert_sufficient_statistics(suff_stats)
        cls_ = cls._projection_class or cls._Base_class or cls
        return cls_.from_natural_parameters(natural_params, **kwargs)

    def __mul__(self, other: Union["AbstractMessage", Real]) -> "AbstractMessage":
        if isinstance(other, MessageInterface):
            return self._multiply(other)
        else:
            cls = self._Base_class or type(self)
            log_norm = self.log_norm + np.log(other)
            return cls(
                *self.parameters,
                log_norm=log_norm,
                id_=self.id,
                lower_limit=self.lower_limit,
                upper_limit=self.upper_limit,
            )

    def __rmul__(self, other: "AbstractMessage") -> "AbstractMessage":
        return self * other

    def __truediv__(self, other: Union["AbstractMessage", Real]) -> "AbstractMessage":
        if isinstance(other, MessageInterface):
            return self._divide(other)
        else:
            cls = self._Base_class or type(self)
            log_norm = self.log_norm - np.log(other)
            return cls(
                *self.parameters,
                log_norm=log_norm,
                id_=self.id,
                lower_limit=self.lower_limit,
                upper_limit=self.upper_limit,
            )

    def __pow__(self, other: Real) -> "AbstractMessage":
        natural = self.natural_parameters
        new_params = other * natural
        log_norm = other * self.log_norm
        new = self.from_natural_parameters(
            new_params,
            log_norm=log_norm,
            id_=self.id,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
        )
        return new

    @classmethod
    def parameter_names(cls):
        return getfullargspec(cls.__init__).args[1:-1]

    def __str__(self) -> str:
        param_attrs = [
            (attr, np.asanyarray(getattr(self, attr)))
            for attr in self.parameter_names()
        ]
        if self.shape:
            pad = max(len(attr) for attr, _ in param_attrs)
            attr_str = "    {:<%d}={}" % pad
            param_strs = ",\n".join(
                attr_str.format(attr, np.array2string(val, prefix=" " * (pad + 5)))
                for attr, val in param_attrs
            )
            return f"{type(self).__name__}(\n{param_strs})"
        else:
            param_strs = ", ".join(
                attr + "=" + np.array2string(val, prefix=" " * (len(attr) + 1))
                for attr, val in param_attrs
            )
            return f"{type(self).__name__}({param_strs})"

    __repr__ = __str__

    def factor(self, x):
        # self.assert_within_limits(x)
        return self.logpdf(x)

    @classmethod
    def project(
        cls, samples: np.ndarray, log_weight_list: Optional[np.ndarray] = None, **kwargs
    ) -> "AbstractMessage":
        """Calculates the sufficient statistics of a set of samples
        and returns the distribution with the appropriate parameters
        that match the sufficient statistics
        """
        # if weight_list aren't passed then equally weight all samples

        # Numerically stable weighting for very small/large weight_list

        # rescale coordinates to 'natural parameter space'
        if log_weight_list is None:
            log_weight_list = np.zeros_like(samples)

        log_w_max = np.max(log_weight_list, axis=0, keepdims=True)
        w = np.exp(log_weight_list - log_w_max)
        norm = w.mean(0)
        log_norm = np.log(norm) + log_w_max[0]

        tx = cls.to_canonical_form(samples)
        w /= norm
        suff_stats = (tx * w[None, ...]).mean(1)

        assert np.isfinite(suff_stats).all()

        cls_ = cls._projection_class or cls._Base_class or cls
        return cls_.from_sufficient_statistics(suff_stats, log_norm=log_norm, **kwargs)

    @classmethod
    def from_mode(
        cls, mode: np.ndarray, covariance: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        pass

    def log_normalisation(self, *elems: Union["AbstractMessage", float]) -> np.ndarray:
        """
        Calculates the log of the integral of the product of a
        set of distributions

        NOTE: ignores log normalisation
        """
        # Remove floats from messages passed

        dists: List[MessageInterface] = [
            dist
            for dist in self._iter_dists(elems)
            if isinstance(dist, MessageInterface,)
        ]

        # Calculate log product of message normalisation
        log_norm = self.log_base_measure - self.log_partition
        log_norm += sum(dist.log_base_measure - dist.log_partition for dist in dists)

        # Calculate log normalisation of product of messages
        prod_dist = self.sum_natural_parameters(*dists)
        log_norm -= prod_dist.log_base_measure - prod_dist.log_partition

        return log_norm

    def instance(self):
        return self

    def update_invalid(self, other: "AbstractMessage") -> "AbstractMessage":
        valid = self.check_valid()
        if self.ndim:
            valid_parameters: Iterator[np.ndarray] = (
                np.where(valid, p, p_safe) for p, p_safe in zip(self, other)
            )
        else:
            # TODO: Fairly certain this would not work
            valid_parameters = iter(self if valid else other)
        cls = self._Base_class or type(self)
        new = cls(
            *valid_parameters,
            log_norm=self.log_norm,
            id_=self.id,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
        )
        return new

    @staticmethod
    def _get_mean_variance(
        mean: np.ndarray, covariance: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        mean, covariance = np.asanyarray(mean), np.asanyarray(covariance)

        if not covariance.shape:
            # If variance is float simply pass through
            variance = covariance * np.ones_like(mean)
            if not variance.shape:
                variance = variance.item()
        elif mean.shape == covariance.shape:
            variance = np.asanyarray(covariance)
        elif covariance.shape == mean.shape * 2:
            # If 2D covariance matrix passed get diagonal
            inds = tuple(np.indices(mean.shape))
            variance = np.asanyarray(covariance)[inds * 2]
        else:
            raise ValueError(
                f"shape of covariance {covariance.shape} is invalid "
                f"must be (), {mean.shape}, or {mean.shape * 2}"
            )
        return mean, variance

    def __call__(self, x):
        return np.sum(self.logpdf(x))

    def factor_jacobian(
        self, x: np.ndarray, _variables: Optional[Tuple[str]] = ("x",)
    ) -> Union[np.ndarray, Tuple[np.ndarray, Tuple[np.ndarray, ...]]]:
        loglike, g = self.logpdf_gradient(x)
        g = np.expand_dims(g, list(range(loglike.ndim)))
        return loglike.sum(), (g,)

    def as_factor(self, variable: "Variable", name: Optional[str] = None):
        from autofit.graphical.factor_graphs import Factor

        if name is None:
            shape = self.shape
            clsname = type(self).__name__
            family = clsname[:-7] if clsname.endswith("Message") else clsname
            name = f"{family}Likelihood" + (str(shape) if shape else "")

        return Factor(
            self,
            variable,
            name=name,
            factor_jacobian=self.factor_jacobian,
            plates=variable.plates,
            arg_names=["x"],
        )

    def calc_exact_update(self, x: "AbstractMessage") -> "AbstractMessage":
        return (self,)

    def has_exact_projection(self, x: "AbstractMessage") -> bool:
        return type(self) is type(x)

    @classmethod
    def _reconstruct(
        cls,
        parameters: Tuple[np.ndarray, ...],
        log_norm: float,
        id_,
        lower_limit,
        upper_limit,
        *args,
    ):
        return cls(
            *parameters,
            log_norm=log_norm,
            id_=id_,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

    def __reduce__(self):
        # serialises TransformedMessage during pickling
        return (
            self._reconstruct,
            (
                self.parameters,
                self.log_norm,
                self.id,
                self.lower_limit,
                self.upper_limit,
            ),
        )

    def _sample(self, n_samples):
        # Needed for nested TransformedMessage method resolution
        return self.sample(n_samples)

    @classmethod
    def _logpdf_gradient(cls, self, x):
        # Needed for nested TransformedMessage method resolution
        return cls.logpdf_gradient(self, x)

    @classmethod
    def _logpdf_gradient_hessian(cls, self, x):
        # Needed for nested TransformedMessage method resolution
        return cls.logpdf_gradient_hessian(self, x)


def map_dists(
    dists: Dict[str, AbstractMessage],
    values: Dict[str, np.ndarray],
    _call: str = "logpdf",
) -> Iterator[Tuple[str, np.ndarray]]:
    """
    Calls a method (default: logpdf) for each Message in dists
    on the corresponding value in values
    """
    for v in dists.keys() & values.keys():
        dist = dists[v]
        if isinstance(dist, MessageInterface):
            yield v, getattr(dist, _call)(values[v])



import math
from typing import Optional, Tuple

import numpy as np

from autoconf import cached_property
from autofit.messages.abstract import AbstractMessage


class FixedMessage(AbstractMessage):
    log_base_measure = 0

    def __init__(
            self,
            value: np.ndarray,
            lower_limit=-math.inf,
            upper_limit=math.inf,
            log_norm: np.ndarray = 0.,
            id_=None
    ):
        self.value = value
        super().__init__(
            value,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            log_norm=log_norm,
            id_=id_
        )

    def value_for(self, unit: float) -> float:
        raise NotImplemented()

    @cached_property
    def natural_parameters(self) -> Tuple[np.ndarray, ...]:
        return self.parameters

    @staticmethod
    def invert_natural_parameters(natural_parameters: np.ndarray
                                  ) -> Tuple[np.ndarray]:
        return natural_parameters,

    @staticmethod
    def to_canonical_form(x: np.ndarray) -> np.ndarray:
        return x

    @cached_property
    def log_partition(self) -> np.ndarray:
        return 0.

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats: np.ndarray
                                     ) -> np.ndarray:
        return suff_stats

    def sample(self, n_samples: Optional[int] = None) -> np.ndarray:
        """
        Rely on array broadcasting to get fixed values to
        calculate correctly
        """
        if n_samples is None:
            return self.value
        return np.array([self.value])

    logpdf_cache = {}

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        if x.shape not in FixedMessage.logpdf_cache:
            FixedMessage.logpdf_cache[x.shape] = np.zeros_like(x)
        return FixedMessage.logpdf_cache[x.shape]

    @cached_property
    def mean(self) -> np.ndarray:
        return self.value

    @cached_property
    def variance(self) -> np.ndarray:
        return np.zeros_like(self.mean)

    def _no_op(self, *other, **kwargs) -> 'FixedMessage':
        """
        'no-op' operation

        In many operations fixed messages should just
        return themselves
        """
        return self

    project = _no_op
    from_mode = _no_op
    __pow__ = _no_op
    __mul__ = _no_op
    __div__ = _no_op
    default = _no_op
    _multiply = _no_op
    _divide = _no_op
    sum_natural_parameters = _no_op
    sub_natural_parameters = _no_op

    def kl(self, dist: "FixedMessage") -> float:
        return 0.
